# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import NoReturn, Optional, Tuple, Union

import torch
from torch import Tensor

from megatron.core import tensor_parallel
from megatron.core.inference.contexts import BaseInferenceContext
from megatron.core.models.common.embeddings.rope_utils import (
    apply_rotary_pos_emb,
    apply_rotary_pos_emb_with_cos_sin,
)
from megatron.core.packed_seq_params import PackedSeqParams
from megatron.core.parallel_state import (
    get_data_parallel_group,
    get_data_parallel_rank,
    get_data_parallel_world_size,
    get_tensor_model_parallel_group,
    get_tensor_model_parallel_rank,
    get_tensor_model_parallel_world_size,
)
from megatron.core.process_groups_config import ProcessGroupCollection
from megatron.core.transformer.identity_op import IdentityOp
from megatron.core.transformer.module import MegatronModule
from megatron.core.transformer.spec_utils import ModuleSpec, build_module
from megatron.core.utils import (
    deprecate_inference_params,
    divide,
    get_pg_size,
    is_fa_min_version,
    is_te_min_version,
    nvtx_range_pop,
    nvtx_range_push,
)

from ..models.common.embeddings.yarn_rotary_pos_embedding import (
    _yarn_get_concentration_factor_from_config,
)
from .enums import AttnMaskType
from .transformer_config import TransformerConfig

try:
    from einops import rearrange
except ImportError:
    rearrange = None

try:
    from flashattn_hopper.flash_attn_interface import _flash_attn_forward
    from flashattn_hopper.flash_attn_interface import (
        flash_attn_with_kvcache as flash_attn3_with_kvcache,
    )

    HAVE_FA3 = True
except:
    HAVE_FA3 = False

try:
    from flash_mla import flash_mla_with_kvcache, get_mla_metadata

    HAVE_FMLA = True
except ImportError:
    flash_mla_with_kvcache = None
    get_mla_metadata = None
    HAVE_FMLA = False

from megatron.core.transformer.transformer_config import MLATransformerConfig

try:
    from flash_attn import flash_attn_varlen_func, flash_attn_with_kvcache
except:
    flash_attn_varlen_func = None
    flash_attn_with_kvcache = None

try:
    import transformer_engine  # pylint: disable=unused-import

    HAVE_TE = True
    from megatron.core.extensions.transformer_engine import (
        SplitAlongDim,
        TELinear,
        set_save_original_input,
    )
except ImportError:
    HAVE_TE = False
    SplitAlongDim, TELinear, set_save_original_input = None, None, None

try:
    from transformer_engine.pytorch.attention.rope import apply_fused_qkv_rotary_pos_emb

    HAVE_FUSED_QKV_ROPE = True
except ImportError:
    HAVE_FUSED_QKV_ROPE = False


@dataclass
class SelfAttentionSubmodules:
    """
    Configuration class for specifying the submodules of a self-attention.
    """

    linear_qkv: Union[ModuleSpec, type] = None
    core_attention: Union[ModuleSpec, type] = None
    linear_proj: Union[ModuleSpec, type] = None
    q_layernorm: Union[ModuleSpec, type] = None
    k_layernorm: Union[ModuleSpec, type] = None


@dataclass
class CrossAttentionSubmodules:
    """
    Configuration class for specifying the submodules of a cross-attention.
    """

    linear_q: Union[ModuleSpec, type] = None
    linear_kv: Union[ModuleSpec, type] = None
    core_attention: Union[ModuleSpec, type] = None
    linear_proj: Union[ModuleSpec, type] = None


class Attention(MegatronModule, ABC):
    """Attention layer abstract class.

    This layer only contains common modules required for the "self attn" and
    "cross attn" specializations.
    """

    def __init__(
        self,
        config: TransformerConfig,
        submodules: Union[SelfAttentionSubmodules, CrossAttentionSubmodules],
        layer_number: int,
        attn_mask_type: AttnMaskType,
        attention_type: str,
        cp_comm_type: str = None,
        pg_collection: ProcessGroupCollection = None,
    ):
        super().__init__(config=config)

        self.config = config
        self.layer_number = layer_number
        self.attn_mask_type = attn_mask_type
        self.attention_type = attention_type

        # For normal attention without groups, num_query_groups == num_attention_heads,
        # so these two will be the same
        self.query_projection_size = self.config.kv_channels * self.config.num_attention_heads
        self.kv_projection_size = self.config.kv_channels * self.config.num_query_groups

        if pg_collection is None:
            pg_collection = ProcessGroupCollection.use_mpu_process_groups(required_pgs=['tp', 'cp'])
        else:
            assert hasattr(
                pg_collection, 'tp'
            ), "Attention pg_collection must have tp process group"
            assert hasattr(
                pg_collection, 'cp'
            ), "Attention pg_collection must have cp process group"
        self.pg_collection = pg_collection

        # Per attention head and per partition values
        world_size = get_pg_size(self.pg_collection.tp)
        self.hidden_size_per_attention_head = divide(
            self.query_projection_size, self.config.num_attention_heads
        )
        self.num_attention_heads_per_partition = divide(self.config.num_attention_heads, world_size)
        self.num_query_groups_per_partition = divide(self.config.num_query_groups, world_size)

        # To support both CUDA Graphs and key value with different hidden size
        self.key_hidden_size = self.hidden_size_per_attention_head
        self.val_hidden_size = self.hidden_size_per_attention_head

        self.core_attention = build_module(
            submodules.core_attention,
            config=self.config,
            layer_number=self.layer_number,
            attn_mask_type=self.attn_mask_type,
            attention_type=self.attention_type,
            cp_comm_type=cp_comm_type,
            softmax_scale=self.config.softmax_scale,
            pg_collection=self.pg_collection,
        )

        self.checkpoint_core_attention = (
            self.config.recompute_granularity == 'selective'
            and "core_attn" in self.config.recompute_modules
        )

        # Output.
        self.linear_proj = build_module(
            submodules.linear_proj,
            self.query_projection_size,
            self.config.hidden_size,
            config=self.config,
            init_method=self.config.output_layer_init_method,
            bias=self.config.add_bias_linear,
            input_is_parallel=True,
            skip_bias_add=True,
            is_expert=False,
            tp_comm_buffer_name='proj',
            tp_group=self.pg_collection.tp,
        )

        if (
            HAVE_TE
            and self.config.fp8
            and self.config.fp8_recipe != 'delayed'
            and is_te_min_version("2.6.0dev0")
            and isinstance(self.linear_proj, TELinear)
        ):
            # For fp8 training, the output of the fused core_attn is saved by itself, and
            # linear_proj also saves the quantized tensor of this output. Here we set the
            # linear_proj to save the original input tensors to avoid the extra memory usage of
            # the quantized tensor.
            set_save_original_input(self.linear_proj)

    def _checkpointed_attention_forward(
        self,
        query,
        key,
        value,
        attention_mask,
        rotary_pos_emb=None,
        attn_mask_type=None,
        attention_bias=None,
        packed_seq_params=None,
    ):
        """Forward method with selective activation checkpointing."""

        def custom_forward(*inputs):
            query = inputs[0]
            key = inputs[1]
            value = inputs[2]
            attention_mask = inputs[3]
            attn_mask_type = inputs[5]
            attn_mask_type = AttnMaskType(attn_mask_type.item())
            output_ = self.core_attention(
                query,
                key,
                value,
                attention_mask,
                attn_mask_type=attn_mask_type,
                attention_bias=attention_bias,
                packed_seq_params=packed_seq_params,
            )
            return output_

        if attn_mask_type is None:
            attn_mask_type = self.attn_mask_type
        attn_mask_type = torch.tensor([attn_mask_type.value], dtype=torch.int)
        hidden_states = tensor_parallel.checkpoint(
            custom_forward, False, query, key, value, attention_mask, rotary_pos_emb, attn_mask_type
        )

        return hidden_states

    def _allocate_memory(self, inference_max_sequence_length, batch_size, dim, dtype):
        """Allocate memory to store kv cache during inference."""

        return torch.empty(
            inference_max_sequence_length,
            batch_size,
            self.num_query_groups_per_partition,
            dim,
            dtype=dtype,
            device=torch.cuda.current_device(),
        )

    def _adjust_key_value_for_inference(
        self,
        inference_context: BaseInferenceContext,
        query: Tensor,
        key: Tensor,
        value: Tensor,
        rotary_pos_emb: Tensor,
        rotary_pos_cos: Optional[Tensor] = None,
        rotary_pos_sin: Optional[Tensor] = None,
        rotary_pos_cos_sin: Optional[Tensor] = None,
        sequence_len_offset: Optional[int] = None,
        *,
        inference_params: Optional[BaseInferenceContext] = None,
    ) -> Tuple[Tensor, Tensor, Tensor, Tensor, Tensor, Tensor]:
        """
        Saves the generated key and value tensors to the end of the buffers in inference_context.
        Returns the full size keys and values from the provided inference_context, as well as
        adjusted rotary_pos_emb.

        Args:
            query (Tensor): Query tensor.
            key (Tensor): Key tensor.
            value (Tensor): Value tensor.
            rotary_pos_emb (Optional[Union[Tensor, Tuple[Tensor, Tensor]]]): Rotary
                embedding tensor(s).
            rotary_pos_cos (Optional[Tensor]): Rotary embedding cosine.
            rotary_pos_sin (Optional[Tensor]): Rotary embedding sine.
            rotary_pos_cos_sin (Optional[Tensor]): Combined rotary embedding cosine and sine.
            Currently used exclusively for inference with dynamic batching and flashinfer RoPE.
            sequence_len_offset (Optional[int]): Sequence length offset used for
                inference CUDA graphs.

        Return:
            Tuple of: query, key, value, rotary_pos_emb, attn_mask_type, block_table.
        """

        inference_context = deprecate_inference_params(inference_context, inference_params)

        attn_mask_type = self.attn_mask_type
        if inference_context is None:
            return query, key, value, rotary_pos_emb, attn_mask_type, None

        # =================================================
        # Pre-allocate memory for key-values for inference.
        # =================================================
        if inference_context.is_static_batching():
            if self.layer_number not in inference_context.key_value_memory_dict:
                inf_max_seq_length = inference_context.max_sequence_length
                inf_max_batch_size = inference_context.max_batch_size
                inference_key_memory = self._allocate_memory(
                    inf_max_seq_length, inf_max_batch_size, self.key_hidden_size, key.dtype
                )
                inference_value_memory = self._allocate_memory(
                    inf_max_seq_length, inf_max_batch_size, self.val_hidden_size, value.dtype
                )
                inference_context.key_value_memory_dict[self.layer_number] = (
                    inference_key_memory,
                    inference_value_memory,
                )
            else:
                # Get the pre-allocated buffers for this layer
                inference_key_memory, inference_value_memory = (
                    inference_context.key_value_memory_dict[self.layer_number]
                )

        if not inference_context.is_static_batching() or inference_context.sequence_len_offset > 0:
            # This should mean that we are past the prompt forward_step
            # and so we need to turn off masking
            attn_mask_type = AttnMaskType.no_mask

        if inference_context.is_static_batching():
            batch_start = inference_context.batch_size_offset
            batch_end = batch_start + key.size(1)
            assert batch_end <= inference_key_memory.size(1)
            sequence_start = inference_context.sequence_len_offset
            sequence_end = sequence_start + key.size(0)
            assert sequence_end <= inference_key_memory.size(0), (
                "Current sequence length is longer than expected maximum sequence length! "
                "Increase inference_max_seq_length."
            )

        if self.config.flash_decode:
            rotary_pos_cos_q = None
            rotary_pos_sin_q = None
            rotary_pos_cos_k = None
            rotary_pos_sin_k = None

            assert inference_context.is_static_batching()
            if (
                inference_context.sequence_len_offset > 0 and rotary_pos_cos is not None
            ):  # Decode phase, not prefill
                rotary_pos_cos_q = rotary_pos_cos[sequence_end - 1 : sequence_end]
                rotary_pos_sin_q = rotary_pos_sin[sequence_end - 1 : sequence_end]
                rotary_pos_cos_k = rotary_pos_cos[sequence_end - 1 : sequence_end]
                rotary_pos_sin_k = rotary_pos_sin[sequence_end - 1 : sequence_end]
            elif rotary_pos_cos is not None:  # Prefill
                rotary_pos_cos_q = rotary_pos_cos[:sequence_end]
                rotary_pos_sin_q = rotary_pos_sin[:sequence_end]
                rotary_pos_cos_k = rotary_pos_cos[:sequence_end]
                rotary_pos_sin_k = rotary_pos_sin[:sequence_end]

            # Flash Decoding assumes that the keys stored in the KV Cache already have RoPE applied.
            # Apply RoPE before we store the keys to make it compatible with flash decoding kernel
            if rotary_pos_sin_q is not None and rotary_pos_sin_k is not None:
                key = apply_rotary_pos_emb_with_cos_sin(
                    key,
                    rotary_pos_cos_k,
                    rotary_pos_sin_k,
                    rotary_interleaved=self.config.rotary_interleaved,
                )
                query = apply_rotary_pos_emb_with_cos_sin(
                    query,
                    rotary_pos_cos_q,
                    rotary_pos_sin_q,
                    rotary_interleaved=self.config.rotary_interleaved,
                )
        else:
            rotary_pos_cos_q = None
            rotary_pos_sin_q = None

        # Adjust rotary embeddings.
        if rotary_pos_emb is not None:
            q_pos_emb, k_pos_emb = rotary_pos_emb
            if inference_context.is_static_batching():
                q_pos_emb = q_pos_emb[sequence_start:sequence_end, :, :, :]
                k_pos_emb = k_pos_emb[:sequence_end, :, :, :]
            else:
                pass
            rotary_pos_emb = (q_pos_emb, k_pos_emb)

        block_table = None
        if inference_context.is_static_batching():
            # Copy key and values.
            inference_key_memory[sequence_start:sequence_end, batch_start:batch_end, ...] = key
            inference_value_memory[sequence_start:sequence_end, batch_start:batch_end, ...] = value
            key = inference_key_memory[:sequence_end, batch_start:batch_end, ...]
            value = inference_value_memory[:sequence_end, batch_start:batch_end, ...]
        else:
            # Apply rotary embeddings before appending KV cache.
            if inference_context.use_flashinfer_fused_rope and (rotary_pos_cos_sin is not None):
                query, key = inference_context.apply_fused_qk_rotary_emb(
                    query, key, rotary_pos_cos_sin, self.config
                )
            elif rotary_pos_emb is not None:
                q_pos_emb, k_pos_emb = rotary_pos_emb
                key = inference_context.apply_rotary_emb_key(
                    key, k_pos_emb, self.config, self.pg_collection.cp
                )

                rotary_pos_emb = (q_pos_emb, None)  # key rotary emb has been applied

            # Append key/value data tensors to cache.
            inference_context.append_key_value_cache(self.layer_number, key, value)

            _, max_seqlen_q = inference_context.cu_query_lengths()
            if getattr(self.config, "cache_mla_latents", None) and max_seqlen_q > 1:
                # Doing unabsorbed MLA Attention with cached mla latents (prefill/mixed mode)
                kv_cache, _, block_table = inference_context.key_value_cache(self.layer_number)
                # Uncompress the KV cache for prefill/mixed mode
                key, value = self.uncompress_kv_from_cache(kv_cache)
            else:
                # Read key/value *pointer* tensors from cache.
                key, value, block_table = inference_context.key_value_cache(self.layer_number)
        return query, key, value, rotary_pos_emb, attn_mask_type, block_table

    @abstractmethod
    def get_query_key_value_tensors(self, hidden_states, key_value_states, split_qkv=True):
        """
        This method needs to be implemented based on whether the derived class
        is "self-attn" or "cross-attn".
        """

    def flash_decode(
        self,
        sequence_len_offset: Tensor,
        query_layer: Tensor,
        key_layer: Tensor,
        value_layer: Tensor,
        inference_key_memory: Tensor,
        inference_value_memory: Tensor,
        rotary_cos: Tensor,
        rotary_sin: Tensor,
        rotary_interleaved: bool = False,
    ) -> (Tensor, Tensor):
        """
        The flash decoding kernel will do the following in a single execution:
        1. Compute RoPE embedding with precomputed cos & sin tensors
        2. Update the KV Cache
        3. Performs the flash attention operation
        """
        assert flash_attn_with_kvcache is not None, (
            "Flash Decoding requires the flash_attn_with_kvcache kernel, "
            "available in the flash-attn package."
        )
        q = query_layer.permute(1, 0, 2, 3)
        k = key_layer.permute(1, 0, 2, 3)
        v = value_layer.permute(1, 0, 2, 3)
        k_cache = inference_key_memory.permute(1, 0, 2, 3)
        v_cache = inference_value_memory.permute(1, 0, 2, 3)

        if rotary_cos is not None:
            rotary_cos = rotary_cos.to(query_layer.dtype)
        if rotary_sin is not None:
            rotary_sin = rotary_sin.to(query_layer.dtype)

        out = flash_attn_with_kvcache(
            q=q,
            k_cache=k_cache,
            v_cache=v_cache,
            k=k,
            v=v,
            rotary_cos=rotary_cos,
            rotary_sin=rotary_sin,
            cache_seqlens=sequence_len_offset,
            rotary_interleaved=rotary_interleaved,
        )
        return out

    def flash_decode_and_prefill(
        self,
        q: Tensor,
        k: Tensor,
        v: Tensor,
        max_seqlen_q,
        max_seqlen_k,
        cu_seqlens_q,
        cu_seqlens_k,
        seqlens_k,
        block_table,
    ) -> Tensor:
        """Flash attention kernel for mixed decode and prefill samples.

        Args:
            q (Tensor): Query tensor.
            k (Tensor): Key tensor.
            v (Tensor): Value tensor.
            max_seqlen_q (int): Query total sequence length.
            max_seqlen_k (int): Key total sequence length.
            cu_seqlens_q (Tensor): Cumulative query sequence lengths.
            cu_seqlens_k (Tensor): Cumulative key sequence lengths.
            seqlens_k (Tensor): key sequence lengths.
            block_table (Tensor): KV cache block ids for all samples.
        Return:
            (Tensor) Attention output.
        """

        assert not self.training
        assert block_table is not None

        # Flash attn kernel.
        if max_seqlen_q > 1:
            q = q.squeeze(1)
            if getattr(self, "softmax_scale", None) is not None:
                softmax_scale = self.softmax_scale
            else:
                softmax_scale = q.shape[-1] ** -0.5
            if HAVE_FA3:
                # TODO(ksanthanam): Replace with call to flash_attn_varlen_func once
                # it accepts block_table
                output_total, *unused = _flash_attn_forward(
                    q=q,
                    k=k,
                    v=v,
                    k_new=None,
                    v_new=None,
                    qv=None,
                    out=None,
                    cu_seqlens_q=cu_seqlens_q,
                    cu_seqlens_k=None,
                    cu_seqlens_k_new=None,
                    seqused_q=None,
                    seqused_k=seqlens_k,
                    max_seqlen_q=max_seqlen_q,
                    max_seqlen_k=max_seqlen_k,
                    page_table=block_table,
                    kv_batch_idx=None,
                    leftpad_k=None,
                    rotary_cos=None,
                    rotary_sin=None,
                    seqlens_rotary=None,
                    q_descale=None,
                    k_descale=None,
                    v_descale=None,
                    softmax_scale=softmax_scale,
                    causal=True,
                    window_size=(-1, -1),
                    attention_chunk=0,
                    softcap=0.0,
                    rotary_interleaved=True,
                    scheduler_metadata=None,
                    num_splits=0,
                    pack_gqa=None,
                    sm_margin=0,
                )
            else:
                output_total = flash_attn_varlen_func(
                    q,
                    k,
                    v,
                    cu_seqlens_q,
                    cu_seqlens_k,
                    max_seqlen_q,
                    max_seqlen_k,
                    softmax_scale=softmax_scale,
                    causal=True,
                    block_table=block_table,
                )
            output_total = output_total.unsqueeze(1)
        else:  # decode only
            # If using MLA we use the FlashMLA kernel
            if isinstance(self.config, MLATransformerConfig):
                softmax_scale = self.softmax_scale

                num_heads_k = 1  # Only a single head for MLA Flash
                seq_len_q = 1  # Sequence length is 1 for decode
                num_heads_q = self.num_attention_heads_per_partition
                num_heads_per_head_k = seq_len_q * num_heads_q // num_heads_k

                cache_seqlens = seqlens_k
                tile_scheduler_metadata, num_splits = get_mla_metadata(
                    cache_seqlens,  # cumulative key-lengths
                    num_heads_per_head_k,  # decode-only lengths
                    num_heads_k,  # per-head dim of V
                )
                head_dim_v = self.config.kv_lora_rank
                kv_cache = k.unsqueeze(-2)
                output_total, softmax_lse = flash_mla_with_kvcache(
                    q,
                    kv_cache,
                    block_table,
                    cache_seqlens,
                    head_dim_v,
                    tile_scheduler_metadata,
                    num_splits,
                    softmax_scale=softmax_scale,
                    causal=True,
                )
            else:
                flash_attn_args = {
                    "q": q,
                    "k_cache": k,
                    "v_cache": v,
                    "cache_seqlens": seqlens_k,
                    "causal": True,
                    "page_table" if HAVE_FA3 else "block_table": block_table,
                }
                if HAVE_FA3:
                    output_total = flash_attn3_with_kvcache(**flash_attn_args)
                else:
                    output_total = flash_attn_with_kvcache(**flash_attn_args)
        return output_total

    def forward(
        self,
        hidden_states: Tensor,
        attention_mask: Tensor,
        key_value_states: Optional[Tensor] = None,
        inference_context: Optional[BaseInferenceContext] = None,
        rotary_pos_emb: Optional[Union[Tensor, Tuple[Tensor, Tensor]]] = None,
        rotary_pos_cos: Optional[Tensor] = None,
        rotary_pos_sin: Optional[Tensor] = None,
        rotary_pos_cos_sin: Optional[Tensor] = None,
        attention_bias: Optional[Tensor] = None,
        packed_seq_params: Optional[PackedSeqParams] = None,
        sequence_len_offset: Optional[int] = None,
        *,
        inference_params: Optional[BaseInferenceContext] = None,
    ) -> Tuple[Tensor, Tensor]:
        """
        Perform a forward pass through the attention module.

        Args:
            hidden_states (Tensor): Hidden states.
            attention_mask (Tensor): Attention mask.
            key_value_states (Optional[Tensor]): Key/value states (for cross attention).
            inference_context (Optional[BaseInferenceContext]): Inference context that manages
                KV cache.
            rotary_pos_emb (Optional[Union[Tensor, Tuple[Tensor, Tensor]]]): Rotary
                embedding tensor(s).
            rotary_pos_cos (Optional[Tensor]): Rotary embedding cosine.
            rotary_pos_sin (Optional[Tensor]): Rotary embedding sine.
            rotary_pos_cos_sin (Optional[Tensor]): Combined rotary embedding cosine and sine.
            Currently used exclusively for inference with dynamic batching and flashinfer RoPE.
            attention_bias (Optional[Tensor]): Attention bias.
            packed_seq_params (Optional[PackedSeqparams]): Parameters used for THD format.
            sequence_len_offset (Optional[int]): Sequence length offset used for
                inference CUDA graphs.

        Return:
            (Tuple[Tensor, Tensor]) Attention output and bias.

        """
        # Check if we need to skip RoPE
        # no_rope is 0-indexed array and self.layer_number is 1-indexed
        no_rope = (
            self.config.no_rope_freq[self.layer_number - 1] if self.config.no_rope_freq else False
        )
        if no_rope:
            rotary_pos_emb = None

        inference_context = deprecate_inference_params(inference_context, inference_params)

        if inference_context and inference_context.is_dynamic_batching():
            assert HAVE_FA3 or is_fa_min_version(
                "2.7.3"
            ), "flash attn verion v2.7.3 and above is required for dynamic batching."

        # hidden_states: [sq, b, h]
        is_inference_mode = inference_context is not None and not self.training
        # is_using_flash_decode - True is we are using the static inference engine with flash decode
        is_using_flash_decode = is_inference_mode and self.config.flash_decode
        # is_using_flashinfer_rope - True if we are using the dynamic inference engine
        # with flashinfer fused rope
        is_using_flashinfer_rope = is_inference_mode and (
            not inference_context.is_static_batching()
            and inference_context.use_flashinfer_fused_rope
        )
        if is_using_flash_decode or is_using_flashinfer_rope:
            # flash decode and flash-infer fused rope use rotary_pos_cos and rotary_pos_sin
            rotary_pos_emb = None
        else:
            assert rotary_pos_cos is None and rotary_pos_sin is None

        # For self attention we just duplicate the rotary_pos_emb if it isn't already
        if rotary_pos_emb is not None and not isinstance(rotary_pos_emb, tuple):
            rotary_pos_emb = (rotary_pos_emb,) * 2

        # =====================
        # Query, Key, and Value
        # =====================
        # Get the query, key and value tensors based on the type of attention -
        # self or cross attn.
        nvtx_range_push(suffix="qkv")
        split_qkv = (self.attention_type == "cross") or not all(
            [
                not self.config.test_mode,
                self.config.fused_single_qkv_rope,
                inference_context is None,
                packed_seq_params is None,
                (
                    rotary_pos_emb is not None
                    and rotary_pos_emb[0] is not None
                    and rotary_pos_emb[1] is not None
                ),
                not self.config.flash_decode,
                HAVE_FUSED_QKV_ROPE,
                self.q_layernorm is None or isinstance(self.q_layernorm, IdentityOp),
                self.k_layernorm is None or isinstance(self.k_layernorm, IdentityOp),
            ]
        )
        # Check if fused_single_qkv_rope is requested but either unavailable or not
        # supported for the current use case.
        if self.attention_type != "cross":
            assert not (
                self.config.fused_single_qkv_rope and split_qkv
            ), "fused_single_qkv_rope requested but not available/supported for the config."

        qkv_output = self.get_query_key_value_tensors(
            hidden_states, key_value_states, split_qkv=split_qkv
        )
        attn_mask_type = self.attn_mask_type
        block_table = None
        if split_qkv:
            query, key, value = qkv_output
        else:
            mixed_qkv, qkv_split_arg_list = qkv_output
        nvtx_range_pop(suffix="qkv")

        # ===================================================
        # Adjust key, value, and rotary_pos_emb for inference
        # ===================================================

        in_decode_mode = (
            inference_context is not None
            and inference_context.is_decode_only()
            and not self.training
        )

        # This branch only runs in the decode phase of flash decoding and returns after the linear
        # projection. This conditional is not used in the prefill phase or non-flash-decoding cases.
        nvtx_range_push(suffix="adjust_key_value")
        if in_decode_mode and self.config.flash_decode:
            assert self.layer_number in inference_context.key_value_memory_dict
            assert inference_context.sequence_len_offset is not None
            inference_key_memory, inference_value_memory = inference_context.key_value_memory_dict[
                self.layer_number
            ]
            output = self.flash_decode(
                sequence_len_offset=sequence_len_offset,
                query_layer=query,
                key_layer=key,
                value_layer=value,
                inference_key_memory=inference_key_memory,
                inference_value_memory=inference_value_memory,
                rotary_cos=rotary_pos_cos,
                rotary_sin=rotary_pos_sin,
                rotary_interleaved=self.config.rotary_interleaved,
            )
            out = output.transpose(0, 1).contiguous()
            context_layer = out.view(out.size(0), out.size(1), -1)
            output, bias = self.linear_proj(context_layer)
            return output, bias

        if (
            in_decode_mode
            and self.config.cuda_graph_impl == "local"
            and self.config.cuda_graph_scope != "full_iteration"
            and inference_context.is_static_batching()
        ):
            raise ValueError(f"CUDA graphs must use flash decode with static batching!")

        if split_qkv:
            query, key, value, rotary_pos_emb, attn_mask_type, block_table = (
                self._adjust_key_value_for_inference(
                    inference_context,
                    query,
                    key,
                    value,
                    rotary_pos_emb,
                    rotary_pos_cos,
                    rotary_pos_sin,
                    rotary_pos_cos_sin,
                    sequence_len_offset,
                )
            )

        if packed_seq_params is not None:
            query = query.squeeze(1)
            key = key.squeeze(1)
            value = value.squeeze(1)
        nvtx_range_pop(suffix="adjust_key_value")

        # ================================================
        # relative positional embedding (rotary embedding)
        # ================================================
        nvtx_range_push(suffix="rotary_pos_emb")
        if rotary_pos_emb is not None and (
            not self.config.flash_decode or inference_context is None
        ):
            q_pos_emb, k_pos_emb = rotary_pos_emb

            if packed_seq_params is not None:
                if packed_seq_params.cu_seqlens_q_padded is not None:
                    cu_seqlens_q = packed_seq_params.cu_seqlens_q_padded
                else:
                    cu_seqlens_q = packed_seq_params.cu_seqlens_q
                if packed_seq_params.cu_seqlens_kv_padded is not None:
                    cu_seqlens_kv = packed_seq_params.cu_seqlens_kv_padded
                else:
                    cu_seqlens_kv = packed_seq_params.cu_seqlens_kv
            else:
                cu_seqlens_q = cu_seqlens_kv = None

            if split_qkv:
                if q_pos_emb is not None:
                    # TODO VIJAY: simplify
                    if inference_context is None or inference_context.is_static_batching():
                        query = apply_rotary_pos_emb(
                            query,
                            q_pos_emb,
                            config=self.config,
                            cu_seqlens=cu_seqlens_q,
                            mscale=_yarn_get_concentration_factor_from_config(self.config),
                            cp_group=self.pg_collection.cp,
                        )
                    else:
                        query = inference_context.apply_rotary_emb_query(
                            query, q_pos_emb, self.config, cu_seqlens_q, self.pg_collection.cp
                        )
                if k_pos_emb is not None:
                    key = apply_rotary_pos_emb(
                        key,
                        k_pos_emb,
                        config=self.config,
                        cu_seqlens=cu_seqlens_kv,
                        mscale=_yarn_get_concentration_factor_from_config(self.config),
                        cp_group=self.pg_collection.cp,
                    )
            else:
                query, key, value = apply_fused_qkv_rotary_pos_emb(
                    mixed_qkv, q_pos_emb, k_pos_emb, qkv_split_arg_list
                )

            # TODO, can apply positional embedding to value_layer so it has
            # absolute positional embedding.
            # otherwise, only relative positional embedding takes effect
            # value_layer = apply_rotary_pos_emb(value_layer, k_pos_emb)
        nvtx_range_pop(suffix="rotary_pos_emb")

        # ==================================
        # core attention computation
        # ==================================

        nvtx_range_push(suffix="core_attention")
        if self.checkpoint_core_attention and self.training:
            core_attn_out = self._checkpointed_attention_forward(
                query,
                key,
                value,
                attention_mask,
                attn_mask_type=attn_mask_type,
                attention_bias=attention_bias,
                packed_seq_params=packed_seq_params,
            )
        else:
            if inference_context is None or inference_context.is_static_batching():
                # Static batching attention kernel.
                core_attn_out = self.core_attention(
                    query,
                    key,
                    value,
                    attention_mask,
                    attn_mask_type=attn_mask_type,
                    attention_bias=attention_bias,
                    packed_seq_params=packed_seq_params,
                )

            else:
                # Dynamic batching attention kernel.
                q, k, v = (query, key, value)
                cu_query_lengths, max_seqlen_q = inference_context.cu_query_lengths()
                cu_kv_lengths, kv_lengths, max_seqlen_k = inference_context.cu_kv_lengths()

                core_attn_out = self.flash_decode_and_prefill(
                    q,
                    k,
                    v,
                    max_seqlen_q,
                    max_seqlen_k,
                    cu_query_lengths,
                    cu_kv_lengths,
                    kv_lengths,
                    block_table,
                )
                core_attn_out = rearrange(core_attn_out, 's b h d -> s b (h d)')

        if packed_seq_params is not None and packed_seq_params.qkv_format == 'thd':
            # reshape to same output shape as unpacked case
            # (t, np, hn) -> (t, b=1, h=np*hn)
            # t is the pack size = sum (sq_i)
            # note that batch is a dummy dimension in the packed case
            core_attn_out = core_attn_out.reshape(core_attn_out.size(0), 1, -1)
        nvtx_range_pop(suffix="core_attention")

        # =================
        # Output. [sq, b, h]
        # =================

        nvtx_range_push(suffix="linear_proj")
        output, bias = self.linear_proj(core_attn_out)
        nvtx_range_pop(suffix="linear_proj")

        return output, bias

    def set_for_recompute_input_layernorm(self):
        """Set the attention layer for recompute input_layernorm. Only needed for fp8."""
        raise NotImplementedError("set_for_recompute_input_layernorm is not implemented.")


class SelfAttention(Attention):
    """Self-attention layer class

    Self-attention layer takes input with size [s, b, h]
    and returns output of the same size.
    """

    def __init__(
        self,
        config: TransformerConfig,
        submodules: SelfAttentionSubmodules,
        layer_number: int,
        attn_mask_type=AttnMaskType.padding,
        cp_comm_type: str = None,
        pg_collection: ProcessGroupCollection = None,
    ):
        super().__init__(
            config=config,
            submodules=submodules,
            layer_number=layer_number,
            attn_mask_type=attn_mask_type,
            attention_type="self",
            cp_comm_type=cp_comm_type,
            pg_collection=pg_collection,
        )

        self.linear_qkv = build_module(
            submodules.linear_qkv,
            self.config.hidden_size,
            self.query_projection_size + 2 * self.kv_projection_size,
            config=self.config,
            init_method=self.config.init_method,
            gather_output=False,
            bias=self.config.add_bias_linear or self.config.add_qkv_bias,
            skip_bias_add=False,
            is_expert=False,
            tp_comm_buffer_name='qkv',
            tp_group=self.pg_collection.tp,
        )

        if submodules.q_layernorm is not None:
            self.q_layernorm = build_module(
                submodules.q_layernorm,
                hidden_size=self.hidden_size_per_attention_head,
                config=self.config,
                eps=self.config.layernorm_epsilon,
            )
        else:
            self.q_layernorm = None

        if submodules.k_layernorm is not None:
            self.k_layernorm = build_module(
                submodules.k_layernorm,
                hidden_size=self.hidden_size_per_attention_head,
                config=self.config,
                eps=self.config.layernorm_epsilon,
            )
        else:
            self.k_layernorm = None

    def run_realtime_tests(self):
        """Performs a consistency check.

        This function makes sure that tensors across devices are the same during an experiment.
        This is often not guaranteed to be so because of silent hardware failures (eg, memory
        corruption loading a checkpoint, network traffic corruption encountered during
        data transmission).

        (TODO) In the future, more tensors should be checked across the training run and
        checked every X iterations. This is left for future work. Equality of tensors is probably
        not required; transmitting hashes is sufficient."""

        if not self.config.qk_layernorm:
            return

        # check that all tensor parallel and data parallel ranks have the same
        # Q & K layernorm parameters.
        rank = get_data_parallel_rank()
        inputs = torch.stack(
            [
                self.q_layernorm.weight.data,
                self.q_layernorm.bias.data,
                self.k_layernorm.weight.data,
                self.k_layernorm.bias.data,
            ]
        )
        dp_list = [torch.empty_like(inputs) for _ in range(get_data_parallel_world_size())]
        dp_list[rank] = inputs
        torch.distributed.all_gather(dp_list, inputs, group=get_data_parallel_group())

        def _compare(srcs, tgts, names, parallelism):
            assert len(srcs) == len(tgts) == len(names)
            for src, tgt, name in zip(srcs, tgts, names):
                assert torch.all(src == tgt), (
                    f"Discrepancy between {name} in {parallelism} ranks {i} and {rank}. "
                    f"Diff: {torch.norm(src - tgt)}"
                )

        for i, dp in enumerate(dp_list):
            q_w, q_b, k_w, k_b = torch.unbind(dp)
            _compare(
                [q_w, q_b, k_w, k_b],
                [
                    self.q_layernorm.weight.data,
                    self.q_layernorm.bias.data,
                    self.k_layernorm.weight.data,
                    self.k_layernorm.bias.data,
                ],
                ["q_w", "q_b", "k_w", "k_b"],
                "DP",
            )

        rank = get_tensor_model_parallel_rank()
        tp_list = [torch.empty_like(inputs) for _ in range(get_tensor_model_parallel_world_size())]
        tp_list[rank] = inputs
        torch.distributed.all_gather(tp_list, inputs, group=get_tensor_model_parallel_group())

        for i, tp in enumerate(tp_list):
            q_w, q_b, k_w, k_b = torch.unbind(tp)
            _compare(
                [q_w, q_b, k_w, k_b],
                [
                    self.q_layernorm.weight.data,
                    self.q_layernorm.bias.data,
                    self.k_layernorm.weight.data,
                    self.k_layernorm.bias.data,
                ],
                ["q_w", "q_b", "k_w", "k_b"],
                "TP",
            )

    def get_query_key_value_tensors(self, hidden_states, key_value_states=None, split_qkv=True):
        """
        Derives `query`, `key` and `value` tensors from `hidden_states`. If `split_qkv=False`, then
        the unsplit mixed_qkv tensor is returned.
        """
        # Attention heads [sq, b, h] --> [sq, b, ng * (np/ng + 2) * hn)]
        mixed_qkv, _ = self.linear_qkv(hidden_states)

        # [sq, b, hp] --> [sq, b, ng, (np/ng + 2) * hn]
        new_tensor_shape = mixed_qkv.size()[:-1] + (
            self.num_query_groups_per_partition,
            (
                (self.num_attention_heads_per_partition // self.num_query_groups_per_partition + 2)
                * self.hidden_size_per_attention_head
            ),
        )
        mixed_qkv = mixed_qkv.view(*new_tensor_shape)

        split_arg_list = [
            (
                self.num_attention_heads_per_partition
                // self.num_query_groups_per_partition
                * self.hidden_size_per_attention_head
            ),
            self.hidden_size_per_attention_head,
            self.hidden_size_per_attention_head,
        ]

        # Return unsplit mixed_qkv and split_arg_list
        if not split_qkv:
            return mixed_qkv, split_arg_list

        if SplitAlongDim is not None:

            # [sq, b, ng, (np/ng + 2) * hn]
            # --> [sq, b, ng, np/ng * hn], [sq, b, ng, hn], [sq, b, ng, hn]
            (query, key, value) = SplitAlongDim(mixed_qkv, 3, split_arg_list)
        else:

            # [sq, b, ng, (np/ng + 2) * hn]
            # --> [sq, b, ng, np/ng * hn], [sq, b, ng, hn], [sq, b, ng, hn]
            (query, key, value) = torch.split(mixed_qkv, split_arg_list, dim=3)

        # [sq, b, ng, np/ng * hn] -> [sq, b, np, hn]
        query = query.reshape(query.size(0), query.size(1), -1, self.hidden_size_per_attention_head)

        if self.q_layernorm is not None:
            query = self.q_layernorm(query)

        if self.k_layernorm is not None:
            key = self.k_layernorm(key)

        if self.config.test_mode:
            self.run_realtime_tests()

        return query, key, value

    def backward_dw(self) -> NoReturn:
        """Execute weight update operations"""
        self._backward_qkv_proj()
        self._backward_output_proj()

    def _backward_qkv_proj(self):
        """Update weights for QKV projection layer"""
        self.linear_qkv.backward_dw()

    def _backward_output_proj(self):
        """Update weights for output projection layer"""
        self.linear_proj.backward_dw()

    def set_for_recompute_input_layernorm(self):
        """Set the attention layer for recompute input_layernorm. Only needed for fp8."""
        from megatron.core.extensions.transformer_engine import set_save_original_input

        set_save_original_input(self.linear_qkv)


class CrossAttention(Attention):
    """Cross-attention layer class

    Cross-attention layer takes input with size [s, b, h] and context with size
    [s, b, h] and returns output of the same size.
    """

    def __init__(
        self,
        config: TransformerConfig,
        submodules: CrossAttentionSubmodules,
        layer_number: int,
        attn_mask_type=AttnMaskType.padding,
        cp_comm_type: str = None,
        pg_collection: ProcessGroupCollection = None,
    ):
        super().__init__(
            config=config,
            submodules=submodules,
            layer_number=layer_number,
            attn_mask_type=attn_mask_type,
            attention_type="cross",
            cp_comm_type=cp_comm_type,
            pg_collection=pg_collection,
        )

        if self.config.num_query_groups != self.config.num_attention_heads:
            raise ValueError("Group query attention is not currently supported in cross attention.")
        assert self.query_projection_size == self.kv_projection_size

        self.linear_q = build_module(
            submodules.linear_q,
            self.config.hidden_size,
            self.query_projection_size,
            config=self.config,
            init_method=self.config.init_method,
            gather_output=False,
            bias=self.config.add_bias_linear,
            skip_bias_add=False,
            is_expert=False,
        )

        self.linear_kv = build_module(
            submodules.linear_kv,
            self.config.hidden_size,
            2 * self.kv_projection_size,
            config=self.config,
            init_method=self.config.init_method,
            gather_output=False,
            bias=self.config.add_bias_linear,
            skip_bias_add=False,
            is_expert=False,
        )

    def get_query_key_value_tensors(self, hidden_states, key_value_states, split_qkv=True):
        """
        Derives `query` tensor from `hidden_states`, and `key`/`value` tensors
        from `key_value_states`.
        """
        assert split_qkv, "split_qkv must be True for CrossAttention"
        # Attention heads [sk, b, h] --> [sk, b, (np * 2 * hn)]
        mixed_kv, _ = self.linear_kv(key_value_states)

        # [sk, b, (np * 2 * hn)] --> [sk, b, np, 2 * hn]
        new_tensor_shape = mixed_kv.size()[:-1] + (
            self.num_attention_heads_per_partition,
            2 * self.hidden_size_per_attention_head,
        )
        mixed_kv = mixed_kv.view(*new_tensor_shape)

        # [sk, b, np, 2 * hn] --> 2 [sk, b, np, hn]
        (key, value) = tensor_parallel.split_tensor_along_last_dim(mixed_kv, 2)

        # Attention head [sq, b, h] --> [sq, b, hp]
        query, _ = self.linear_q(hidden_states)

        # [sq, b, hp] --> [sq, b, np, hn]
        new_tensor_shape = query.size()[:-1] + (
            self.num_attention_heads_per_partition,
            self.hidden_size_per_attention_head,
        )
        query = query.view(*new_tensor_shape)

        return query, key, value



# Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.

"""Tokenization utilities."""


import torch

from megatron.core import parallel_state
from megatron.core.inference.communication_utils import broadcast_int_list, broadcast_tensor


def tokenize_prompts(
    tokenizer, prompts=None, tokens_to_generate=None, add_BOS=None, rank=0, data_parallel=False
):
    """Tokenize prompts and make them avaiable on all ranks.

    Args:
        data_parallel (bool): Broadcast tokens across a single data parallel model replica.
    """

    # On all ranks set to None so we can pass them to functions
    sizes_list = None
    prompts_tokens_cuda_long_tensor = None
    prompts_length_cuda_long_tensor = None

    # On the specified rank, build the above.
    src_rank = torch.distributed.get_rank()
    if data_parallel:
        src_rank = parallel_state.get_data_parallel_src_rank()

    if src_rank == rank:
        assert prompts is not None
        assert tokens_to_generate is not None
        # Tensor of tokens padded and their unpadded length.
        prompts_tokens_cuda_long_tensor, prompts_length_cuda_long_tensor = (
            _tokenize_prompts_and_batch(tokenizer, prompts, tokens_to_generate, add_BOS)
        )
        # We need the sizes of these tensors for the boradcast
        sizes_list = [
            prompts_tokens_cuda_long_tensor.size(0),  # Batch size
            prompts_tokens_cuda_long_tensor.size(1),
        ]  # Sequence lenght

    # First, broadcast the sizes.
    sizes_tensor = broadcast_int_list(
        2, int_list=sizes_list, rank=rank, data_parallel=data_parallel
    )

    # Now that we have the sizes, we can boradcast the tokens
    # and length tensors.
    sizes = sizes_tensor.tolist()
    prompts_tokens_cuda_long_tensor = broadcast_tensor(
        sizes,
        torch.int64,
        tensor=prompts_tokens_cuda_long_tensor,
        rank=rank,
        data_parallel=data_parallel,
    )
    prompts_length_cuda_long_tensor = broadcast_tensor(
        sizes[0],
        torch.int64,
        tensor=prompts_length_cuda_long_tensor,
        rank=rank,
        data_parallel=data_parallel,
    )

    return prompts_tokens_cuda_long_tensor, prompts_length_cuda_long_tensor


def _tokenize_prompts_and_batch(tokenizer, prompts, tokens_to_generate, add_BOS):
    """Given a set of prompts and number of tokens to generate:
    - tokenize prompts
    - set the sequence length to be the max of length of prompts
      plus the number of tokens we would like to generate
    - pad all the sequences to this length so we can convert them
      into a 2D tensor.
    """

    # Tokenize all the prompts.
    if hasattr(tokenizer, 'eod'):
        eod_token = tokenizer.eod
    elif hasattr(tokenizer, 'eos_id'):
        eod_token = tokenizer.eos_id
    else:
        raise AttributeError('No eod token found in Tokenizer')
    if add_BOS:
        prompts_tokens = [[eod_token] + tokenizer.tokenize(prompt) for prompt in prompts]
    else:
        prompts_tokens = [tokenizer.tokenize(prompt) for prompt in prompts]

    # Now we have a list of list of tokens which each list has a different
    # size. We want to extend this list to:
    #   - incorporate the tokens that need to be generated
    #   - make all the sequences equal length.
    # Get the prompts length.
    prompts_length = [len(prompt_tokens) for prompt_tokens in prompts_tokens]
    # Get the max prompts length.
    max_prompt_len = max(prompts_length)
    # Number of tokens in the each sample of the batch.
    samples_length = max_prompt_len + tokens_to_generate
    # Now update the list of list to be of the same size: samples_length.
    for prompt_tokens, prompt_length in zip(prompts_tokens, prompts_length):
        padding_size = samples_length - prompt_length
        prompt_tokens.extend([eod_token] * padding_size)

    # Now we are in a structured format, we can convert to tensors.
    prompts_tokens_tensor = torch.tensor(prompts_tokens, dtype=torch.long, device='cuda')
    prompts_length_tensor = torch.tensor(prompts_length, dtype=torch.long, device='cuda')

    return prompts_tokens_tensor, prompts_length_tensor



# Copyright (c) 2022, NVIDIA CORPORATION. All rights reserved.
from typing import Optional, Union

import torch
import torch.nn as nn

from megatron.core.transformer.enums import AttnMaskType
from megatron.core.transformer.utils import get_default_causal_mask, get_sliding_window_causal_mask


class ScaledUpperTriangMaskedSoftmax(torch.autograd.Function):
    """
    Fused operation which performs following three operations in sequence
    1. Scale the tensor.
    2. Apply upper triangular mask (typically used in gpt models).
    3. Perform softmax.
    """

    @staticmethod
    def forward(ctx, inputs, scale):
        """Forward pass for scaled upper-triangular masked softmax.

        Args:
            ctx: Autograd context used to stash tensors for backward.
            inputs (torch.Tensor): Input tensor of shape [attn_batches, sq, sk].
            scale (float): Scaling factor applied prior to softmax.

        Returns:
            torch.Tensor: Softmax results after applying scale and causal upper-triangular mask.
        """
        import scaled_upper_triang_masked_softmax_cuda

        scale_t = torch.tensor([scale])
        softmax_results = scaled_upper_triang_masked_softmax_cuda.forward(inputs, scale_t[0])

        ctx.save_for_backward(softmax_results, scale_t)
        return softmax_results

    @staticmethod
    def backward(ctx, output_grads):
        """Backward pass for scaled upper-triangular masked softmax.

        Args:
            ctx: Autograd context containing saved tensors from forward.
            output_grads (torch.Tensor): Upstream gradients matching forward output shape.

        Returns:
            Tuple[torch.Tensor, None]: Gradient with respect to inputs and None for scale.
        """
        import scaled_upper_triang_masked_softmax_cuda

        softmax_results, scale_t = ctx.saved_tensors
        input_grads = scaled_upper_triang_masked_softmax_cuda.backward(
            output_grads, softmax_results, scale_t[0]
        )

        return input_grads, None


class ScaledMaskedSoftmax(torch.autograd.Function):
    """
    Fused operation which performs following three operations in sequence
    1. Scale the tensor.
    2. Apply the mask.
    3. Perform softmax.
    """

    @staticmethod
    def forward(ctx, inputs, mask, scale):
        """Forward pass for scaled masked softmax.

        Args:
            ctx: Autograd context used to stash tensors for backward.
            inputs (torch.Tensor): Input tensor of shape [b, np, sq, sk].
            mask (torch.Tensor): Additive mask broadcastable to inputs.
            scale (float): Scaling factor applied prior to softmax.

        Returns:
            torch.Tensor: Softmax results after applying scale and mask.
        """
        import scaled_masked_softmax_cuda

        scale_t = torch.tensor([scale])

        softmax_results = scaled_masked_softmax_cuda.forward(inputs, mask, scale_t[0])
        ctx.save_for_backward(softmax_results, scale_t)
        return softmax_results

    @staticmethod
    def backward(ctx, output_grads):
        """Backward pass for scaled masked softmax.

        Args:
            ctx: Autograd context containing saved tensors from forward.
            output_grads (torch.Tensor): Upstream gradients matching forward output shape.

        Returns:
            Tuple[torch.Tensor, None, None]: Gradient w.r.t inputs; None for mask and scale.
        """
        import scaled_masked_softmax_cuda

        softmax_results, scale_t = ctx.saved_tensors

        input_grads = scaled_masked_softmax_cuda.backward(output_grads, softmax_results, scale_t[0])
        return input_grads, None, None


class ScaledSoftmax(torch.autograd.Function):
    """
    Fused operation which performs following two operations in sequence
    1. Scale the tensor.
    2. Perform softmax.
    """

    @staticmethod
    def forward(ctx, inputs, scale):
        """Forward pass for scaled softmax (no mask).

        Args:
            ctx: Autograd context used to stash tensors for backward.
            inputs (torch.Tensor): Input tensor of shape [b, np, sq, sk] or [attn_batches, sq, sk].
            scale (float): Scaling factor applied prior to softmax.

        Returns:
            torch.Tensor: Softmax results after applying scale.
        """
        import scaled_softmax_cuda

        scale_t = torch.tensor([scale])

        softmax_results = scaled_softmax_cuda.forward(inputs, scale_t[0])
        ctx.save_for_backward(softmax_results, scale_t)
        return softmax_results

    @staticmethod
    def backward(ctx, output_grads):
        """Backward pass for scaled softmax (no mask).

        Args:
            ctx: Autograd context containing saved tensors from forward.
            output_grads (torch.Tensor): Upstream gradients matching forward output shape.

        Returns:
            Tuple[torch.Tensor, None, None]: Gradient w.r.t inputs; None for unused args.
        """
        import scaled_softmax_cuda

        softmax_results, scale_t = ctx.saved_tensors

        input_grads = scaled_softmax_cuda.backward(output_grads, softmax_results, scale_t[0])
        return input_grads, None, None


class SoftmaxOne(nn.Module):
    r"""
    Softmax-off-by-one function as introduced in
    https://www.evanmiller.org/attention-is-off-by-one.html
    Supports fixed or learnable offset
    """

    def __init__(
        self, dim: Optional[int] = None, denominator_offset: Union[torch.Tensor, float] = 1.0
    ) -> None:
        super().__init__()
        self.dim = dim
        self.denominator_offset = denominator_offset

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """forward pass"""
        # sink: [np] --> [1, np, 1, 1] --> [b, np, sq, 1]
        sink = self.denominator_offset.reshape(1, -1, 1, 1).expand(x.size(0), -1, x.size(2), -1)
        # qk: [b, np, sq, sk] --> [b, np, sq, sk+1]
        qk = torch.cat([x, sink], dim=-1)
        # do softmax, and remove sink token at the end
        ret = torch.softmax(qk, dim=-1)[..., :-1]
        return ret


class FusedScaleMaskSoftmax(nn.Module):
    """
    fused operation: scaling + mask + softmax

    Args:
        input_in_fp16: flag to indicate if input in fp16 data format.
        input_in_bf16: flag to indicate if input in bf16 data format.
        attn_mask_type: attention mask type (pad or causal)
        scaled_masked_softmax_fusion: flag to indicate user want to use softmax fusion
        mask_func: mask function to be applied.
        softmax_in_fp32: if true, softmax in performed at fp32 precision.
        scale: scaling factor used in input tensor scaling.
    """

    def __init__(
        self,
        input_in_fp16,
        input_in_bf16,
        attn_mask_type,
        scaled_masked_softmax_fusion,
        mask_func,
        softmax_in_fp32,
        scale,
        window_size=None,
    ):
        super(FusedScaleMaskSoftmax, self).__init__()
        self.input_in_fp16 = input_in_fp16
        self.input_in_bf16 = input_in_bf16
        assert not (
            self.input_in_fp16 and self.input_in_bf16
        ), "both fp16 and bf16 flags cannot be active at the same time."
        self.input_in_float16 = self.input_in_fp16 or self.input_in_bf16
        self.attn_mask_type = attn_mask_type
        self.scaled_masked_softmax_fusion = scaled_masked_softmax_fusion
        self.mask_func = mask_func
        self.softmax_in_fp32 = softmax_in_fp32
        self.scale = scale
        self.window_size = window_size
        assert self.scale is None or softmax_in_fp32, "softmax should be in fp32 when scaled"

    def forward(
        self,
        input: torch.Tensor,
        mask: Optional[torch.Tensor],
        softmax_offset: Optional[torch.Tensor] = None,
    ):
        """Forward pass of softmax with masked input.

        In case attn_mask_type is causal the mask is generated and None can be passed.
        A user-defined mask is only needed when attn_mask_type is not causal.
        """
        # [b, np, sq, sk]
        assert input.dim() == 4

        if self.is_kernel_available(mask, *input.size()) and softmax_offset is None:
            return self.forward_fused_softmax(input, mask)
        else:
            return self.forward_torch_softmax(input, mask, softmax_offset)

    def is_kernel_available(self, mask, b, np, sq, sk):
        """Check whether the fused CUDA kernel can be used for the given shapes and settings.

        Args:
            mask (Optional[torch.Tensor]): Attention mask or None.
            b (int): Batch size.
            np (int): Number of attention heads per tensor-parallel partition.
            sq (int): Query sequence length.
            sk (int): Key sequence length.

        Returns:
            bool: True if the fused kernel constraints are satisfied; otherwise False.
        """
        attn_batches = b * np

        if (
            self.scaled_masked_softmax_fusion  # user want to fuse
            and self.input_in_float16  # input must be fp16
            and 16 < sk <= 4096  # sk must be 16 ~ 2048
            and sq % 4 == 0  # sq must be divisor of 4
            and sk % 4 == 0  # sk must be divisor of 4
            and attn_batches % 4 == 0  # np * b must be divisor of 4
        ):
            if 0 <= sk <= 4096:
                batch_per_block = self.get_batch_per_block(sq, sk, b, np)

                if self.attn_mask_type == AttnMaskType.causal:
                    if attn_batches % batch_per_block == 0:
                        return True
                else:
                    if sq % batch_per_block == 0:
                        return True
        return False

    def forward_fused_softmax(self, input, mask):
        """Compute softmax using fused CUDA kernels when available.

        Args:
            input (torch.Tensor): Attention scores of shape [b, np, sq, sk].
            mask (Optional[torch.Tensor]): Optional mask for non-causal attention.

        Returns:
            torch.Tensor: Attention probabilities of shape [b, np, sq, sk].
        """
        b, np, sq, sk = input.size()
        scale = self.scale if self.scale is not None else 1.0

        if self.attn_mask_type == AttnMaskType.causal:
            assert sq == sk, "causal mask is only for self attention"

            # input is 3D tensor (attn_batches, sq, sk)
            input = input.view(-1, sq, sk)
            probs = ScaledUpperTriangMaskedSoftmax.apply(input, scale)
            return probs.view(b, np, sq, sk)
        else:
            # input is 4D tensor (b, np, sq, sk)
            if mask is not None:
                return ScaledMaskedSoftmax.apply(input, mask, scale)
            else:
                return ScaledSoftmax.apply(input, scale)

    def forward_torch_softmax(self, input, mask, softmax_offset=None):
        """Fallback PyTorch implementation for masked softmax.

        Applies optional scaling, constructs a causal or sliding-window mask if needed,
        applies the mask, and computes softmax in PyTorch. Optionally casts back to
        float16/bfloat16 when requested.

        Args:
            input (torch.Tensor): Attention scores of shape [b, np, sq, sk].
            mask (Optional[torch.Tensor]): Optional additive mask.

        Returns:
            torch.Tensor: Attention probabilities of shape [b, np, sq, sk].
        """
        if self.input_in_float16 and self.softmax_in_fp32:
            input = input.float()

        if self.scale is not None:
            input = input * self.scale

        # Generate causal mask if not given
        sq, sk = input.size(2), input.size(3)
        if self.window_size is not None:
            mask = get_sliding_window_causal_mask(sq, sk, self.window_size)
        elif self.attn_mask_type == AttnMaskType.causal and mask is None and sq > 1:
            # If sq == 1 then either KV cache is used or one-element context is passed
            # so keeping mask=None in this case; subsequent code should handle it
            assert sq == sk, "causal mask is only for self attention"
            mask = get_default_causal_mask(sq)

        mask_output = self.mask_func(input, mask) if mask is not None else input
        if softmax_offset is None:
            softmax_fn = torch.nn.Softmax(dim=-1)
        else:
            softmax_fn = SoftmaxOne(-1, softmax_offset.to(input.device))

        probs = softmax_fn(mask_output)
        if self.input_in_float16 and self.softmax_in_fp32:
            if self.input_in_fp16:
                probs = probs.half()
            else:
                probs = probs.bfloat16()

        return probs

    @staticmethod
    def get_batch_per_block(sq, sk, b, np):
        """Return CUDA kernel's batch-per-block parameter for masked softmax.

        Args:
            sq (int): Query sequence length.
            sk (int): Key sequence length.
            b (int): Batch size.
            np (int): Number of attention heads per tensor-parallel partition.

        Returns:
            int: Batch-per-block value as computed by the CUDA extension.
        """
        import scaled_masked_softmax_cuda

        return scaled_masked_softmax_cuda.get_batch_per_block(sq, sk, b, np)



# Copyright (c) 2024, NVIDIA CORPORATION. All rights reserved.
import json
from typing import Union

import torch

from megatron.core.export.data_type import DataType
from megatron.core.export.export_config import ExportConfig
from megatron.core.export.model_type import ModelType
from megatron.core.export.trtllm.engine_builder.trtllm_engine_builder import TRTLLMEngineBuilder
from megatron.core.export.trtllm.model_to_trllm_mapping.default_conversion_dict import (
    DEFAULT_CONVERSION_DICT,
    NEMOTRON_NAS_CONVERSION_DICT,
)
from megatron.core.export.trtllm.trt_model_config import TRT_MODEL_CONFIG
from megatron.core.export.trtllm.trt_model_type import TRT_MODEL_TYPE_STRING
from megatron.core.export.trtllm.trtllm_layers import TRTLLMLayers

# pylint: disable=line-too-long
from megatron.core.export.trtllm.trtllm_weights_converter.distributed_trtllm_model_weights_converter import (
    DistributedTRTLLMModelWeightsConverter,
)
from megatron.core.export.trtllm.trtllm_weights_converter.single_device_trtllm_model_weights_converter import (
    SingleDeviceTRTLLMModelWeightsConverter,
)
from megatron.core.export.trtllm.trtllm_weights_converter.utils import is_gated_activation
from megatron.core.transformer.transformer_config import TransformerConfig

try:
    import tensorrt_llm
    from tensorrt_llm.functional import non_gated_version
    from tensorrt_llm.layers import MoeConfig

    HAVE_TRTLLM = True
except ImportError:
    HAVE_TRTLLM = False


class TRTLLMHelper:
    """TRTLLM Helper class to convert export and build TRTLLM model."""

    def __init__(
        self,
        *,
        transformer_config: TransformerConfig,
        model_type: ModelType,
        trtllm_conversion_dict: dict = {},
        position_embedding_type: str = "learned_absolute",
        max_position_embeddings: int = None,
        rotary_percentage: int = 1.0,
        rotary_base: int = 10000,
        rope_scaling_factor: float = 8.0,
        moe_tp_mode: int = 2,
        multi_query_mode: bool = False,
        activation: str = "gelu",
        seq_len_interpolation_factor: float = None,
        moe_renorm_mode=None,
        share_embeddings_and_output_weights=False,
    ):
        """Constructor for the TRTLLMHelper

        There are two public API's supported  by this helper.
        a) get_trtllm_pretrained_config_and_model_weights
        b) build_and_save_engine

        Args:
            transformer_config (TransformerConfig): The transformer config
            model_type (ModelType): The type of the input model. Enum (megatron.core.export.model_type.ModelType)
            trtllm_conversion_dict (dict, optional): A conversion dictionary that will map your model layer names to trtllm equivalent layer names. Default dictionary is given megatron/core/export/model_to_trtllm_mapping. This dict is merged into the default dict. NOTE: Ignore layer numbers in the model layer names. (e.g) decoder.layers.0.attention_qkv.weight will be decoder.layers.attention_qkv.weight in the mapping dictionary. Defaults to {}.
            position_embedding_type (str, optional): The position embedding type. Defaults to None.
            max_position_embeddings (int, optional): Max posistion embeddings value. Defaults to None.
            rotary_percentage (int, optional): The rotary percentage if using rope embedding. Defaults to 1.0.
            rotary_base (int, optional): The rotary base (theta value) if using rope embeddings. Defaults to 10000.
            moe_tp_mode (int, optional): TRTLLM Config. Defaults to 2.
            multi_query_mode (bool, optional): Defaults to False.
            activation (str, optional): Defaults to "gelu".
            seq_len_interpolation_factor (float, optional): The sequence length interpolation factor if using rope embeddings. Defaults to None.
            moe_renorm_mode (optional) : Renormalization mode if using mixture of experts. Defaults to None.
            share_embeddings_and_output_weights (bool, optional): True if input and output layers share weights. Defaults to False.
        """

        if not HAVE_TRTLLM:
            raise ImportError(
                "tensorrt_llm is not installed. Please install it with `pip install tensorrt-llm`"
            )

        self.transformer_config = transformer_config
        self.model_type = model_type
        self.trtllm_conversion_dict = DEFAULT_CONVERSION_DICT.copy()
        if model_type == ModelType.nemotron_nas:
            self.trtllm_conversion_dict.update(NEMOTRON_NAS_CONVERSION_DICT)
        self.trtllm_conversion_dict.update(trtllm_conversion_dict)
        assert position_embedding_type in [
            "learned_absolute",
            "rope",
        ], f"Position embedding type should be one of learned_absolute, rope. You entered {position_embedding_type}"
        self.position_embedding_type = position_embedding_type
        self.max_position_embeddings = max_position_embeddings
        self.rotary_percentage = rotary_percentage
        self.rotary_base = rotary_base
        self.rope_scaling_factor = rope_scaling_factor
        self.moe_tp_mode = moe_tp_mode
        self.multi_query_mode = multi_query_mode
        self.activation = activation
        self.seq_len_interpolation_factor = seq_len_interpolation_factor
        self.moe_renorm_mode = moe_renorm_mode
        self.share_embeddings_and_output_weights = share_embeddings_and_output_weights
        self.weights_converter = None

    def _get_trtllm_config(
        self,
        export_config: ExportConfig,
        world_size: int,
        gpus_per_node: int,
        vocab_size_padded: int,
        dtype: DataType,
        fp8_quantized: bool = False,
        fp8_kvcache: bool = False,
    ):
        """Get TRTLLM Config

        Returns appropriate TRTLLM PretrainedConfig used by TRTLLM for building engine

        Args:
            export_config (ExportConfig): The export config that defines inference tp , pp size etc.
            world_size (int): The number of gpus (Mostly TP * PP)
            gpus_per_node (int): Num gpus per node
            vocab_size_padded (int): Padded vocab size
            dtype (DataType): The datatype or model precision

        Returns:
            GPTConfig or the LLamaConfig or the PretrainedConfig constructed from your model config
        """
        hidden_act = self.activation
        hidden_act = (
            hidden_act.split("-")[-1]
            if self.transformer_config.num_moe_experts
            else non_gated_version(hidden_act)
        )

        config = {
            "architecture": TRT_MODEL_TYPE_STRING[self.model_type],
            "dtype": dtype.name,
            "num_hidden_layers": self.transformer_config.num_layers,
            "num_attention_heads": self.transformer_config.num_attention_heads,
            "num_key_value_heads": (
                self.transformer_config.num_query_groups
                if self.transformer_config.num_query_groups
                else self.transformer_config.num_attention_heads
            ),
            "head_size": self.transformer_config.kv_channels,
            "hidden_size": self.transformer_config.hidden_size,
            "intermediate_size": self.transformer_config.ffn_hidden_size,
            "norm_epsilon": self.transformer_config.layernorm_epsilon,
            "vocab_size": vocab_size_padded,
            "position_embedding_type": (
                "rope_gpt_neox" if self.position_embedding_type == "rope" else "learned_absolute"
            ),
            "max_position_embeddings": self.max_position_embeddings,
            "hidden_act": hidden_act,
            "use_parallel_embedding": export_config.use_parallel_embedding,
            "embedding_sharding_dim": 0,
            "share_embedding_table": self.share_embeddings_and_output_weights,
            "quantization": {
                "quant_algo": "FP8" if fp8_quantized else None,
                "kv_cache_quant_algo": "FP8" if fp8_kvcache else None,
            },
            "bias": self.transformer_config.add_bias_linear,
            "apply_query_key_layer_scaling": False,
            "rotary_pct": self.rotary_percentage,
            "rotary_base": self.rotary_base,
            "moe_num_experts": (
                0
                if self.transformer_config.moe_router_topk == 0
                else (self.transformer_config.num_moe_experts or 1)
            ),
            "moe_top_k": self.transformer_config.moe_router_topk,
            "moe_normalization_mode": self.moe_renorm_mode
            or MoeConfig.ExpertScaleNormalizationMode.RENORMALIZE,
            "moe_tp_mode": self.moe_tp_mode,
            "logits_dtype": "float32",
            "world_size": world_size,
            "tp_size": export_config.inference_tp_size,
            "pp_size": export_config.inference_pp_size,
            "gpus_per_node": gpus_per_node,
        }

        if self.model_type == ModelType.falcon:
            config["new_decoder_architecture"] = (
                False if self.transformer_config.num_layers == 32 else True
            )
            config["parallel_attention"] = True

        if self.seq_len_interpolation_factor is not None:
            config["rotary_scaling"] = {
                "type": "linear",
                "factor": float(self.seq_len_interpolation_factor),
            }

        if self.model_type == ModelType.nemotron_nas:
            hf_config_dict = json.loads(
                self.transformer_config.heterogeneous_layers_config_encoded_json
            )
            config["block_configs"] = hf_config_dict["block_configs"]
            config["rotary_scaling"] = {"type": "llama3", "factor": self.rope_scaling_factor}

        config_cls = TRT_MODEL_CONFIG[self.model_type]
        return config_cls(**config)

    def _load_scaling_factors(self, model_state_dict: dict) -> dict:
        """Loads scaling factors from model state dictionary.

        Args:
            model_state_dict (dict): Model state dictionary
        Returns:
            dict: Maps scaling factor key, to its value and the inverse. The inverse is used for casting the quantized weights.
        """
        weight_scaling_suffix = ".weights_scaling_factor"
        activation_scaling_suffix = ".activation_scaling_factor"
        mock_scales_dict = {}
        extra_state_infix = "._extra_state"
        mock_suffix = ".weight"

        for key, val in model_state_dict.items():
            if extra_state_infix in key and not key.endswith("core_attention._extra_state"):
                mock_key = key.split(extra_state_infix)[0] + mock_suffix
                mock_scales_dict[mock_key] = val

        mock_scales_dict = TRTLLMLayers.rename_input_layer_names_to_trtllm_layer_names(
            mock_scales_dict, self.trtllm_conversion_dict, False
        )
        split_gated_activation = is_gated_activation(self)

        scales = {}
        for key, val in mock_scales_dict.items():
            if val is None:
                continue

            val.seek(0)
            extra_states = torch.load(val)

            activation_scaling_factor_key = key.replace(mock_suffix, activation_scaling_suffix)
            weight_scaling_factor_key = key.replace(mock_suffix, weight_scaling_suffix)

            activation_scales = {
                "trt_llm_scale": extra_states["scale_inv_fwd"][0].view(1),
                "weight_multiplier": extra_states["scale_fwd"][0].view(1),
            }

            weight_scales = {
                "trt_llm_scale": extra_states["scale_inv_fwd"][1].view(1),
                "weight_multiplier": extra_states["scale_fwd"][1].view(1),
            }

            scales[activation_scaling_factor_key] = activation_scales
            scales[weight_scaling_factor_key] = weight_scales
            if split_gated_activation and ".mlp.fc" in key:
                scales[activation_scaling_factor_key.replace("fc", "gate")] = activation_scales
                scales[weight_scaling_factor_key.replace("fc", "gate")] = weight_scales

        return scales

    # pylint: disable=line-too-long
    def get_trtllm_pretrained_config_and_model_weights(
        self,
        model_state_dict,
        dtype: DataType,
        export_config: ExportConfig = None,
        on_device_distributed_conversion: bool = False,
        vocab_size: int = None,
        gpus_per_node: int = None,
        state_dict_split_by_layer_numbers: bool = True,
        fp8_quantized: bool = False,
        fp8_kvcache: bool = False,
    ):
        """Get TRTLLM Config and Converted Model Weights

        This function returns the trtllm model weights as a list.
        There are two modes for conversion. The default is to use a single device cpu/gpu for conversion.
        NOTE: For faster performance, if your entire model will fit in memory, pre transfer the model state dict to cuda device and then call this function.
        For on device conversion it returns weights which will be used on the device itself.
        Same thing happens with the pretrained config

        Args:
            model_state_dict (dict): The input model state dictionary (Entire model state loaded on CPU) or the model state dict of each GPU in the case of on_device conversion)
            export_config (ExportConfig): The export config used to define inference tp size, pp size etc. Used only for on device conversion.
            dtype (DataType): The data type of model precision
            on_device_distributed_conversion (bool, optional): Convert on gpus in distributed setting. This assumes that the model state dict is sharded according to required inference model parallelism and that each gpu gets its part of the model state dict . Defaults to False.
            vocab_size (int, optional): The vocabulary size. Defaults to None.
            gpus_per_node (int, optional): The number of gpus per node. Used for on device conversion.
            state_dict_split_by_layer_numbers (bool, optional): Are the model layers split by layer numbers in state dict. For example : mlp.fc1.weight can be represented like mlp.fc1.weight of shape [num_layers, hidden_dim, ffn_hidden_dim]} or it can be like mlp.fc1.layers.0.weight of shape [hidden_dim, ffn_hidden_dim], then mlp.fc1.layers.1.weight ... for all layers. If you use represenation 2 set this to True. Defaults to True

        Returns:
            Two lists . First list of trtllm converted model weights(Either on device, or a list of weights for each gpu) and the trtllm_model_configs.
        """
        assert model_state_dict is not None, "Model state dict is not set"

        scales = self._load_scaling_factors(model_state_dict) if fp8_quantized else {}
        model_state_dict = {k: v for k, v in model_state_dict.items() if "extra_state" not in k}

        if on_device_distributed_conversion:
            assert vocab_size is not None, "Need to pass in vocab_size for on device"
            supported_model = self.model_type in [
                ModelType.gpt,
                ModelType.gptnext,
                ModelType.llama,
                ModelType.nemotron_nas,
            ]
            assert (
                supported_model
            ), "On device conversion only supported for model types gptnext and llama"
            assert export_config is None, (
                "Export config is inferred based on the parallel state. "
                "If you want to set inference tp 2, then load the model with this TP2 setting and just pass in the model state dict."
            )

            assert (
                gpus_per_node is not None
            ), "Need to pass in gpus_per_node for on device conversion"
            trtllm_model_weights_on_device, trtllm_model_config = (
                self._get_trtllm_pretrained_config_and_model_weights_in_distributed_setting(
                    model_state_dict,
                    dtype,
                    vocab_size,
                    gpus_per_node,
                    scales,
                    fp8_quantized,
                    fp8_kvcache,
                )
            )
            return [trtllm_model_weights_on_device], [trtllm_model_config]

        else:
            assert (
                vocab_size is None
            ), "Vocab size is inferred from the input layer for cpu conversion. So leave it as None"
            trtllm_model_weights_list, trtllm_model_config_list = (
                self._get_trtllm_pretrained_config_and_model_weights_list_on_single_device(
                    export_config,
                    model_state_dict,
                    dtype,
                    gpus_per_node,
                    state_dict_split_by_layer_numbers,
                    scales,
                    fp8_quantized,
                    fp8_kvcache,
                )
            )

            return trtllm_model_weights_list, trtllm_model_config_list

    def _add_scales_to_converter(
        self,
        converter: Union[
            SingleDeviceTRTLLMModelWeightsConverter, DistributedTRTLLMModelWeightsConverter
        ],
        scales: dict,
        fp8_kvcache: bool,
    ):
        """Adds scaling factors to the distributed and single device converters.

        Args:
            converter (ModelWeightConverter): Converter, holding the TRT-LLM model weights.
            scales (dict): Dictionary holding TRT-LLM scaling factors
            fp8_kvcache (bool): If true, creates scaling factors (equal to 1.0) for kv_cache quantization
        """
        trt_scales = {key: scale["trt_llm_scale"] for key, scale in scales.items()}
        kv_scales = {}
        if fp8_kvcache:
            for key in converter.trtllm_model_weights:
                if ".attention.qkv.weight" in key:
                    kv_key = key.split(".qkv")[0] + ".kv_cache_scaling_factor"
                    kv_scales[kv_key] = torch.tensor([1.0], dtype=torch.float32)

        converter.trtllm_model_weights |= trt_scales | kv_scales

    def _get_trtllm_pretrained_config_and_model_weights_in_distributed_setting(
        self,
        model_state_dict: dict,
        dtype: DataType,
        vocab_size: int,
        gpus_per_node: int,
        scales: dict,
        fp8_quantized: bool,
        fp8_kvcache: bool,
    ):
        """Get the TRTLLM Pretrained config and model weights list in a distributed setting

        This function assumes the  model state dict is distributed according to model parallelism .
        Each device gets its own model state dict

        Args:
            export_config (ExportConfig): The export config to set inference tp, pp size etc.
            model_state_dict (dict): The model state dictionary (All collected on cpu)
            dtype (DataType): The data type or model precision
            vocab_size (int): Tokenizer vocab size
            gpus_per_node (int): The number of gpus per node
            scales (dict): Dictionary with fp8 scaling factors
            fp8_quantized (bool): True for fp8 checkpoint export
            fp8_kvcache (bool): True for fp8 KV-cache quantization
        Returns:
            Two lists . List of trtllm converted model weights and trtllm model configs (One for each gpu).
        """

        self.weights_converter = DistributedTRTLLMModelWeightsConverter(
            transformer_config=self.transformer_config,
            dtype=dtype,
            multi_query_mode=self.multi_query_mode,
            activation=self.activation,
            scales=scales,
        )
        self.weights_converter.convert(
            model_state_dict=model_state_dict,
            trtllm_conversion_dict=self.trtllm_conversion_dict,
            tokenizer_vocab_size=vocab_size,
        )
        self._add_scales_to_converter(self.weights_converter, scales, fp8_kvcache)

        export_config = ExportConfig(
            inference_pp_size=self.weights_converter.inference_pp_size,
            inference_tp_size=self.weights_converter.inference_tp_size,
            use_parallel_embedding=True,
        )

        world_size = export_config.inference_tp_size * export_config.inference_pp_size

        trtllm_model_config = self._get_trtllm_config(
            export_config=export_config,
            world_size=world_size,
            gpus_per_node=gpus_per_node,
            vocab_size_padded=vocab_size,
            dtype=dtype,
            fp8_quantized=fp8_quantized,
            fp8_kvcache=fp8_kvcache,
        )

        model_parallel_rank = (
            self.weights_converter.pp_rank * self.weights_converter.inference_tp_size
            + self.weights_converter.tp_rank
        )

        trtllm_model_config.mapping = tensorrt_llm.Mapping(
            world_size=world_size,
            rank=model_parallel_rank,
            tp_size=export_config.inference_tp_size,
            pp_size=export_config.inference_pp_size,
        )

        return self.weights_converter.trtllm_model_weights, trtllm_model_config

    def _get_trtllm_pretrained_config_and_model_weights_list_on_single_device(
        self,
        export_config: ExportConfig,
        model_state_dict: dict,
        dtype: DataType,
        gpus_per_node,
        state_dict_split_by_layer_numbers,
        scales: dict,
        fp8_quantized: bool,
        fp8_kvcache: bool,
    ):
        """Get the TRTLLM Pretrained config and model weights list (one per gpu rank) on single device (CPU/GPU)

        This function assumes the entire model state dict is present in CPU or on one GPU

        Args:
            export_config (ExportConfig): The export config to set inference tp, pp size etc.
            model_state_dict (dict): The model state dictionary (All collected on cpu)
            dtype (DataType): The data type or model precision
            gpus_per_node (int, optional): Number of gpus per node
            state_dict_split_by_layer_numbers (bool, optional): Are the model layers split by layer numbers in state dict. For example : mlp.fc1.weight can be represented like mlp.fc1.weight of shape [num_layers, hidden_dim, ffn_hidden_dim]} or it can be like mlp.fc1.layers.0.weight of shape [hidden_dim, ffn_hidden_dim], then mlp.fc1.layers.1.weight ... for all layers. If you use represenation 2 set this to True. Defaults to True
            scales (dict): Dictionary with fp8 scaling factors
            fp8_quantized (bool): True for fp8 checkpoint export
            fp8_kvcache (bool): True for fp8 KV-cache quantization

        Returns:
            Two lists . List of trtllm converted model weights and trtllm model configs (One for each gpu).
        """
        trtllm_model_configs_list = []
        trtllm_model_weights_list = []

        self.weights_converter = SingleDeviceTRTLLMModelWeightsConverter(
            export_config=export_config,
            transformer_config=self.transformer_config,
            dtype=dtype,
            activation=self.activation,
            multi_query_mode=self.multi_query_mode,
            scales=scales,
        )
        # Convert the input model state dict to trtllm model weights dictionary
        self.weights_converter.convert(
            model_state_dict=model_state_dict,
            trtllm_conversion_dict=self.trtllm_conversion_dict,
            state_dict_split_by_layer_numbers=state_dict_split_by_layer_numbers,
        )

        self._add_scales_to_converter(self.weights_converter, scales, fp8_kvcache)

        vocab_size_padded = self.weights_converter.get_padded_vocab_size()
        world_size = export_config.inference_tp_size * export_config.inference_pp_size
        gpus_per_node = gpus_per_node or export_config.inference_tp_size

        for gpu_rank in range(world_size):
            mapping = tensorrt_llm.Mapping(
                world_size=world_size,
                rank=gpu_rank,
                tp_size=export_config.inference_tp_size,
                pp_size=export_config.inference_pp_size,
            )

            # Important to create a new instance everytime so that the list elements have differnt rank values in the mapping object
            trtllm_model_config = self._get_trtllm_config(
                export_config=export_config,
                world_size=world_size,
                gpus_per_node=gpus_per_node,
                vocab_size_padded=vocab_size_padded,
                dtype=dtype,
                fp8_quantized=fp8_quantized,
                fp8_kvcache=fp8_kvcache,
            )
            trtllm_model_config.mapping = mapping
            trtllm_model_configs_list.append(trtllm_model_config)

            # Get the model weights for each rank and append it to the trtllm_model_weights_list
            trtllm_model_weights_per_gpu = self.weights_converter.get_local_model_weights_per_gpu(
                mapping, trtllm_model_config
            )
            trtllm_model_weights_list.append(trtllm_model_weights_per_gpu)

        return trtllm_model_weights_list, trtllm_model_configs_list

    def build_and_save_engine(
        self,
        engine_dir: str,
        trtllm_model_weights: dict,
        trtllm_model_config,
        max_input_len: int = 1024,
        max_output_len: int = 1024,
        max_batch_size: int = 4,
        lora_ckpt_list=None,
        use_lora_plugin=None,
        max_lora_rank: int = 64,
        lora_target_modules=None,
        max_prompt_embedding_table_size: int = 0,
        paged_kv_cache: bool = True,
        remove_input_padding: bool = True,
        paged_context_fmha: bool = False,
        use_refit: bool = False,
        max_num_tokens: int = None,
        max_seq_len: int = None,
        opt_num_tokens: int = None,
        max_beam_width: int = 1,
        tokens_per_block: int = 128,
        multiple_profiles: bool = False,
        gpt_attention_plugin: str = "auto",
        gemm_plugin: str = "auto",
    ):
        """Method to build the TRTLLM Engine

        This method uses the TRTLLMEngineBuilder to build and save the engine to engine dir

        Args:
            engine_dir (str): The file path to save the engine
            trtllm_model_weights (dict): The TRTLLM converted model weights dict
            trtllm_model_config : The TRTLLM Config
            max_input_len (int, optional): Max input length. Defaults to 1024.
            max_output_len (int, optional): Max output length. Defaults to 1024.
            max_batch_size (int, optional): Max batch size. Defaults to 4.
            lora_ckpt_list (_type_, optional): Lora checkpoint list. Defaults to None.
            use_lora_plugin (_type_, optional): Use lora plugin. Defaults to None.
            max_lora_rank (int, optional): Max lora rank. Defaults to 64.
            lora_target_modules (_type_, optional): Lora target modules. Defaults to None.
            max_prompt_embedding_table_size (int, optional): Max size of prompt embedding table. Defaults to 0.
            paged_kv_cache (bool, optional): Use Paged KV cache. Defaults to True.
            remove_input_padding (bool, optional): Remove input padding. Defaults to True.
            paged_context_fmha (bool, optional): Paged context fmha. Defaults to False.
            use_refit (bool, optional): Use refit. Defaults to False.
            max_num_tokens (int, optional): Max num of tokens. Defaults to None.
            max_seq_len (int, optional): Max seq length. Defaults to None.
            opt_num_tokens (int, optional): Opt number of tokens. Defaults to None.
            max_beam_width (int, optional): Max beam width. Defaults to 1.
            tokens_per_block (int, optional): Nmber of tokens per block. Defaults to 128.
            multiple_profiles (bool, optional): Use multiple profiles. Defaults to False.
            gpt_attention_plugin (str, optional): Gpt attention plugin to use. Defaults to "auto".
            gemm_plugin (str, optional): Gemma plugin to use. Defaults to "auto".
        """

        engine = TRTLLMEngineBuilder.build_and_save_engine(
            engine_dir,
            trtllm_model_weights,
            trtllm_model_config,
            max_input_len,
            max_output_len,
            max_batch_size,
            lora_ckpt_list,
            use_lora_plugin,
            max_lora_rank,
            lora_target_modules,
            max_prompt_embedding_table_size,
            paged_kv_cache,
            remove_input_padding,
            paged_context_fmha,
            use_refit,
            max_num_tokens,
            max_seq_len,
            opt_num_tokens,
            max_beam_width,
            tokens_per_block,
            multiple_profiles,
            gpt_attention_plugin,
            gemm_plugin,
        )

        return engine



# Copyright (c) 2022-2023, NVIDIA CORPORATION.  All rights reserved.

"""Strategies using TensorStore to load and save Zarr arrays."""

from functools import partial
from itertools import starmap
from logging import getLogger
from pathlib import Path
from typing import Union

import torch

from ..core import CheckpointingException
from ..dict_utils import dict_list_map_inplace
from ..mapping import ShardedStateDict, ShardedTensor
from .base import LoadShardedStrategy, StrategyAction, register_default_strategy
from .zarr import load_zarr_based_sharded_metadata, postprocess_numpy_array

try:
    import tensorstore as ts

    HAVE_TENSORSTORE = True
except ImportError:
    from unittest.mock import MagicMock

    ts = MagicMock()
    HAVE_TENSORSTORE = False


logger = getLogger(__name__)


def register_default_tensorstore_strategies():
    """Register default strategies leveraging tensorstore."""
    register_default_strategy(
        StrategyAction.LOAD_SHARDED, "zarr", 1, TensorStoreLoadShardedStrategy()
    )


class TensorStoreLoadShardedStrategy(LoadShardedStrategy):
    """Load strategy for Zarr backend using `tensorstore` for loading."""

    def __init__(self, load_directly_on_device: bool = False):
        super().__init__()
        self.load_directly_on_device = load_directly_on_device

    def load(self, sharded_state_dict: ShardedStateDict, checkpoint_dir: Union[str, Path]):
        if isinstance(checkpoint_dir, str):
            checkpoint_dir = Path(checkpoint_dir)

        if torch.distributed.get_rank() == 0:
            print(f"Loading distributed checkpoint with {self.__class__.__name__}")
            if self.load_directly_on_device:
                print(f"Loading distributed checkpoint directly on the GPU")
        load_fn = partial(
            _load_from_array,
            checkpoint_dir=checkpoint_dir,
            load_directly_on_device=self.load_directly_on_device,
        )
        dict_list_map_inplace(load_fn, sharded_state_dict)
        return sharded_state_dict

    def load_tensors_metadata(self, checkpoint_dir: Union[str, Path]):
        if isinstance(checkpoint_dir, str):
            checkpoint_dir = Path(checkpoint_dir)

        def get_ts_shape_dtype(path):
            arr = open_ts_array(path)
            return arr.shape, arr.dtype.numpy_dtype

        return load_zarr_based_sharded_metadata(checkpoint_dir, get_ts_shape_dtype)

    def check_backend_compatibility(self, loaded_version):
        pass  # TODO

    def check_version_compatibility(self, loaded_version):
        pass  # TODO


def merge_global_slice_with_shape(global_slice, actual_shape, key):
    """Intersects the global slice with the actual shape (prevent overflow)."""

    def _merge_slice(dim_slice, dim_size):
        if isinstance(dim_slice, slice):
            assert (
                dim_slice.start < dim_size
            ), f"Got empty slice for ShardedTensor {key} ({dim_slice}, {dim_size})"
            if dim_slice.stop > dim_size:
                dim_slice = slice(dim_slice.start, dim_size, dim_slice.step)
        return dim_slice

    assert len(global_slice) == len(actual_shape), (global_slice, actual_shape, key)
    return tuple(starmap(_merge_slice, zip(global_slice, actual_shape)))


def _load_from_array(
    sharded_tensor: ShardedTensor,
    checkpoint_dir: Path,
    load_directly_on_device: bool = False,
    apply_flattened_range: bool = True,
):
    x = _load_regular_chunk(sharded_tensor, checkpoint_dir)
    ten = postprocess_numpy_array(x, sharded_tensor, apply_flattened_range)
    if load_directly_on_device:
        sharded_tensor.data.data.copy_(ten)
        return sharded_tensor.data
    else:
        return ten


def _load_regular_chunk(sharded_tensor: ShardedTensor, checkpoint_dir: Path):
    assert isinstance(sharded_tensor, ShardedTensor), type(sharded_tensor)
    arr = open_ts_array(checkpoint_dir / sharded_tensor.key)
    if sharded_tensor.global_shape == arr.shape:
        x = (
            arr[sharded_tensor.global_slice()].read().result()
        )  # flattened tensors loading is delayed
    elif sharded_tensor.allow_shape_mismatch:
        global_slice = merge_global_slice_with_shape(
            sharded_tensor.global_slice(), arr.shape, sharded_tensor.key
        )
        x = arr[global_slice].read().result()  # flattened tensors loading is delayed
    else:
        _msg = (
            f"Global shape mismatch for loaded ({arr.shape})"
            f" and expected ({sharded_tensor.global_shape}) tensor"
            f" for key {sharded_tensor.key}"
        )
        raise CheckpointingException(_msg)
    return x


def open_ts_array(arr_path: Path):
    """Opens a Zarr file array with Tensorstore with basic setting.

    Args:
        arr_path (Path): path to a Zarr (Tensorstore) array
    """
    if not HAVE_TENSORSTORE:
        raise RuntimeError(
            "tensorstore is required, please install it with `pip install tensorstore`"
        )
    spec = {"driver": "zarr", "metadata_key": ".zarray", "kvstore": {}}
    spec["kvstore"] = {"driver": "file", "path": str(arr_path)}
    try:
        arr = ts.open(ts.Spec(spec), open=True).result()
    except Exception as e:
        raise CheckpointingException(f"Array {arr_path} could not be loaded. Error: {e}") from e
    return arr



# Copyright (c) 2024, NVIDIA CORPORATION. All rights reserved.
import warnings

from megatron.core.fusions.fused_bias_dropout import get_bias_dropout_add
from megatron.core.tensor_parallel.layers import ColumnParallelLinear, RowParallelLinear
from megatron.core.transformer.attention import SelfAttention, SelfAttentionSubmodules
from megatron.core.transformer.dot_product_attention import DotProductAttention
from megatron.core.transformer.enums import AttnMaskType
from megatron.core.transformer.identity_op import IdentityOp
from megatron.core.transformer.mlp import MLP, MLPSubmodules
from megatron.core.transformer.spec_utils import ModuleSpec
from megatron.core.transformer.transformer_layer import TransformerLayer, TransformerLayerSubmodules

try:
    import transformer_engine as te  # pylint: disable=unused-import

    from megatron.core.extensions.transformer_engine import (
        TEDotProductAttention,
        TELayerNormColumnParallelLinear,
        TERowParallelLinear,
    )

    HAVE_TE = True
except ImportError:
    HAVE_TE = False

try:
    import apex  # pylint: disable=unused-import

    from megatron.core.fusions.fused_layer_norm import FusedLayerNorm

    HAVE_APEX = True
    LNImpl = FusedLayerNorm
except ImportError:
    from megatron.core.transformer.torch_norm import WrappedTorchNorm

    warnings.warn("Apex is not installed. Falling back to Torch Norm")
    LNImpl = WrappedTorchNorm
    HAVE_APEX = False


def get_bert_layer_with_transformer_engine_spec():
    """Use this spec to use lower-level Transformer Engine modules (required for fp8 training).

    Returns:
        ModuleSpec: Module specification with TE modules
    """
    if not HAVE_TE:
        raise ImportError(
            "Transformer Engine is not installed. Please use local Bert layer spec instead."
        )

    return ModuleSpec(
        module=TransformerLayer,
        submodules=TransformerLayerSubmodules(
            self_attention=ModuleSpec(
                module=SelfAttention,
                params={"attn_mask_type": AttnMaskType.padding},
                submodules=SelfAttentionSubmodules(
                    linear_qkv=TELayerNormColumnParallelLinear,
                    core_attention=TEDotProductAttention,
                    linear_proj=TERowParallelLinear,
                    q_layernorm=IdentityOp,
                    k_layernorm=IdentityOp,
                ),
            ),
            self_attn_bda=get_bias_dropout_add,
            mlp=ModuleSpec(
                module=MLP,
                submodules=MLPSubmodules(
                    linear_fc1=TELayerNormColumnParallelLinear, linear_fc2=TERowParallelLinear
                ),
            ),
            mlp_bda=get_bias_dropout_add,
        ),
    )


def __getattr__(name):
    if name == "bert_layer_with_transformer_engine_spec":
        warnings.warn(
            """Attribute bert_layer_specs.bert_layer_with_transformer_engine_spec is on a
            deprecation track and will be removed in future releases. Please migrate to
            bert_layer_specs.get_bert_layer_with_transformer_engine_spec()."""
        )

        return get_bert_layer_with_transformer_engine_spec()


# Use this spec for an implementation using only modules in megatron core
bert_layer_local_spec = ModuleSpec(
    module=TransformerLayer,
    submodules=TransformerLayerSubmodules(
        input_layernorm=LNImpl,
        self_attention=ModuleSpec(
            module=SelfAttention,
            params={"attn_mask_type": AttnMaskType.padding},
            submodules=SelfAttentionSubmodules(
                linear_qkv=ColumnParallelLinear,
                core_attention=DotProductAttention,
                linear_proj=RowParallelLinear,
                q_layernorm=IdentityOp,
                k_layernorm=IdentityOp,
            ),
        ),
        self_attn_bda=get_bias_dropout_add,
        pre_mlp_layernorm=LNImpl,
        mlp=ModuleSpec(
            module=MLP,
            submodules=MLPSubmodules(linear_fc1=ColumnParallelLinear, linear_fc2=RowParallelLinear),
        ),
        mlp_bda=get_bias_dropout_add,
        sharded_state_dict_keys_map={
            "input_layernorm.": "self_attention.linear_qkv.layer_norm_",
            "pre_mlp_layernorm.": "mlp.linear_fc1.layer_norm_",
        },
    ),
)



# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.

import json
from dataclasses import asdict, dataclass, field, fields
from pathlib import Path

from megatron.core.transformer import TransformerConfig


@dataclass
class AttentionConfig:
    """Configuration parameters for the self-attention part of a single transformer
    block in a heterogeneous transformer."""

    no_op: bool = False
    """Whether this is a no-op operation."""

    replace_with_linear: bool = False
    """Whether to replace the self-attention mechanism with a single linear layer."""

    num_query_groups: int | None = None
    """Number of query groups for grouped query attention."""

    @classmethod
    def build_config_from_dict(
        cls, block_config_dict: dict, num_attention_heads: int
    ) -> 'AttentionConfig':
        """
        Builds an AttentionConfig object from a dictionary and the number of attention heads.

        Args:
            block_config_dict (dict): The dictionary containing the configuration for the attention.
            num_attention_heads (int): The number of attention heads.

        Returns:
            AttentionConfig: The AttentionConfig object.
        """
        attention_config_dict = block_config_dict["attention"]
        if "num_query_groups" not in attention_config_dict:
            # compatibility with HF config of nvidia/Llama-3_1-Nemotron-51B-Instruct
            n_heads_in_group = attention_config_dict.pop("n_heads_in_group")
            if n_heads_in_group is not None:
                if num_attention_heads % n_heads_in_group != 0:
                    raise ValueError(
                        f"num_attention_heads ({num_attention_heads}) must be a multiple of "
                        f"n_heads_in_group ({n_heads_in_group})."
                    )
                num_query_groups = num_attention_heads // n_heads_in_group
            else:
                num_query_groups = None
            attention_config_dict["num_query_groups"] = num_query_groups

        # keep only fields from cls
        field_names = {f.name for f in fields(cls)}
        attn_config_dict = {k: v for k, v in attention_config_dict.items() if k in field_names}
        return cls(**attn_config_dict)


@dataclass
class MLPConfig:
    """Configuration parameters for the MLP part of a single transformer
    block in a heterogeneous transformer."""

    no_op: bool = False
    """Whether this is a no-op operation."""

    replace_with_linear: bool = False
    """Whether to replace the MLP with a single linear layer."""

    ffn_hidden_size: float | None = None
    """MLP intermediate size"""

    @classmethod
    def build_config_from_dict(cls, block_config_dict: dict, hidden_size: int) -> 'MLPConfig':
        """
        Builds an MLPConfig object from a dictionary and a hidden size.

        Args:
            block_config_dict (dict): The dictionary containing the configuration for the MLP.
            hidden_size (int): The hidden size of the MLP.

        Returns:
            MLPConfig: The constructed MLPConfig object.
        """
        mlp_config_dict = block_config_dict.get("ffn") or block_config_dict.get("mlp")
        if "ffn_hidden_size" not in mlp_config_dict:
            # compatibility with HF config of nvidia/Llama-3_1-Nemotron-51B-Instruct
            ffn_mult = mlp_config_dict.pop("ffn_mult")
            if ffn_mult is not None:
                ffn_hidden_size = cls.ffn_mult_to_intermediate_size(ffn_mult, hidden_size)
            else:
                ffn_hidden_size = None
            mlp_config_dict["ffn_hidden_size"] = ffn_hidden_size

        # keep only fields from cls
        field_names = {f.name for f in fields(cls)}
        mlp_config_dict = {k: v for k, v in mlp_config_dict.items() if k in field_names}
        return cls(**mlp_config_dict)

    @staticmethod
    def ffn_mult_to_intermediate_size(ffn_mult: float, hidden_size: int) -> int:
        """
        Calculates the intermediate size of the MLP based on the given
        `ffn_mult` and `hidden_size`.

        Args:
            ffn_mult (float): The multiplier for the feed-forward network.
            hidden_size (int): The size of the hidden layer.

        Returns:
            int: The calculated intermediate size.
        """
        intermediate_size = int(2 * ffn_mult * hidden_size / 3)
        return MLPConfig.find_multiple(intermediate_size, 256)

    @staticmethod
    def find_multiple(n: int, k: int) -> int:
        """
        Calculates the smallest multiple of `k` greater than or equal to `n`.

        Args:
            n (int): The number to find the multiple of.
            k (int): The number to find the multiple of.

        Returns:
            int: The smallest multiple of `k` greater than or equal to `n`.
        """
        if n % k == 0:
            return n
        return n + k - (n % k)


@dataclass
class TransformerBlockConfig:
    """Configuration parameters for a single transformer block in a heterogeneous transformer."""

    attention: AttentionConfig
    """Configuration parameters for the self-attention part of the transformer block in a 
    heterogeneous transformer."""

    mlp: MLPConfig
    """Configuration parameters for the mlp part of the transformer block in a 
    heterogeneous transformer."""


@dataclass
class HeterogeneousTransformerConfig(TransformerConfig):
    """Configuration object for megatron-core heterogeneous transformers.

    Heterogeneous models refer to transformer architectures where individual layers can differ
    in configuration. Specifically:
        - Attention or MLP layers can be replaced with either a linear layer or a no-op
        - MLP intermediate dimensions can vary between layers
    We use the format of the HuggingFace config files in llama nemotron models to define
    the architecture.
    For example,
    https://huggingface.co/nvidia/Llama-3_3-Nemotron-Super-49B-v1/resolve/main/config.json

    Most notably, the "heterogeneous_layers_config_path" maps to a json file containing a
    "block_configs" key, which is a list of attention and mlp configurations for each layer.
    For example, the "block_config" for a 2 layer model is:
    "block_configs": [
        {
            "attention": {
                "n_heads_in_group": 8,
                "no_op": false,
                "replace_with_linear": false,
            },
            "ffn": {
                "ffn_mult": 2.625,
                "no_op": false,
                "replace_with_linear": false,
            }
        },
        {
            "attention": {
                "n_heads_in_group": null,
                "no_op": true,
                "replace_with_linear": false,
            },
            "ffn": {
                "ffn_mult": 2.625,
                "no_op": false,
                "replace_with_linear": false,
            }
        }
    ]
    """

    heterogeneous_layers_config_path: str = ""
    """Path to the json file containing the heterogeneous block specs."""

    heterogeneous_layers_config_encoded_json: str = ""
    """The contents of the json file containing the heterogeneous block specs. It will be read from 
    heterogeneous_layers_config_path at first, then saved forever inside the model checkpoint."""

    per_block_parameters: list[TransformerBlockConfig] = field(init=False)
    """Configuration parameters for each of the transformer blocks in a 
    heterogeneous transformer."""

    def __post_init__(self):
        super().__post_init__()

        self.heterogeneous_block_specs = True

        if self.heterogeneous_layers_config_encoded_json in ("", None):
            self.heterogeneous_layers_config_encoded_json = Path(
                self.heterogeneous_layers_config_path
            ).read_text()

        hf_config_dict = json.loads(self.heterogeneous_layers_config_encoded_json)
        assert "block_configs" in hf_config_dict
        block_list = hf_config_dict["block_configs"]

        block_configs = [
            TransformerBlockConfig(
                attention=AttentionConfig.build_config_from_dict(
                    block_config_dict=block, num_attention_heads=self.num_attention_heads
                ),
                mlp=MLPConfig.build_config_from_dict(
                    block_config_dict=block, hidden_size=self.hidden_size
                ),
            )
            for block in block_list
        ]

        self.per_block_parameters = block_configs

    def get_config_for_layer(self, layer_number: int) -> TransformerConfig:
        """
        Get the config for the given layer number.
        Based on the layer number, the corresponding block config is returned,
        overriding the main config's value.
        """
        layer_idx = layer_number - 1  # layer number starts from 1
        if layer_idx < 0 or layer_idx >= len(self.per_block_parameters):
            raise ValueError(
                f"Invalid layer number: {layer_number}. Should be in "
                f"range [1, {len(self.per_block_parameters)}]."
            )
        block_config = self.per_block_parameters[layer_idx]

        keys_to_update = {}

        # attention config updates
        if block_config.attention.num_query_groups is not None:
            assert (
                not block_config.attention.replace_with_linear and not block_config.attention.no_op
            )
            keys_to_update['num_query_groups'] = block_config.attention.num_query_groups

        # mlp config updates
        if block_config.mlp.ffn_hidden_size is not None:
            assert not block_config.mlp.replace_with_linear and not block_config.mlp.no_op
            keys_to_update['ffn_hidden_size'] = block_config.mlp.ffn_hidden_size

        transformer_config_dict = asdict(self)

        # remove keys that are not in TransformerConfig
        transformer_config_field_names = {f.name for f in fields(TransformerConfig)}
        transformer_config_dict = {
            k: v for k, v in transformer_config_dict.items() if k in transformer_config_field_names
        }

        transformer_config_dict.update(keys_to_update)

        return TransformerConfig(**transformer_config_dict)



# Copyright (c) 2024, NVIDIA CORPORATION. All rights reserved.
"""Integrations with NVIDIA TensorRT Model Optimizer (referred as ModelOpt).

ModelOpt is a library comprising state-of-the-art model optimization techniques
including quantization and sparsity to compress model for efficient inference on
NVIDIA GPUs. ModelOpt is integrated with Megatron-core to provide a seamless
experience for users to optimize their Megatron-core models for inference.
More details on ModelOpt including installation and usage can be found at
https://github.com/NVIDIA/TensorRT-Model-Optimizer.
"""



# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.

import os
import warnings
from pathlib import Path

from torch.cuda.memory import CUDAPluggableAllocator
from torch.utils.cpp_extension import CUDA_HOME, load_inline

from megatron.core.utils import is_torch_min_version

try:
    if is_torch_min_version("2.8.0"):
        from torch.cuda.memory import MemPool
    else:
        from torch.cuda import MemPool
    _has_mem_pool = True
except ImportError:
    _has_mem_pool = False

_mempool_c_src = r"""
#include <cuda_runtime_api.h>
#include <cstddef>

#define EXPORT extern "C"

EXPORT void* managed_malloc(size_t size, int device, void* stream) {
  (void)stream;
  int cur = -1;
  cudaGetDevice(&cur);
  if (device != cur && device >= 0) cudaSetDevice(device);

  // cudaMallocManaged allows for more memory to be allocated than the device memory size.
  // The cudaMemAttachGlobal flag makes the memory accessible from both host and device.
  void* ptr = nullptr;
  cudaError_t err = cudaMallocManaged(&ptr, (size_t)size, cudaMemAttachGlobal);
  if (err != cudaSuccess) return nullptr;

  if (device >= 0) {
    // cudaMemAdviseSetPreferredLocation sets the preferred location for the memory.
    // This is a hint that tries to prevent data from being migrated away from the device.
    cudaMemAdvise(ptr, (size_t)size, cudaMemAdviseSetPreferredLocation, device);
    // cudaMemAdviseSetAccessedBy ensures the memory always lives in the device's page table.
    // Even if the memory has to be migrated away from the device, it still does not page fault.
    // The CUDA docs claim that cudaMemAdviseSetPreferredLocation completely overrides this flag,
    // but there is no harm in adding this flag as well for future-proofing.
    cudaMemAdvise(ptr, (size_t)size, cudaMemAdviseSetAccessedBy, device);
  }
  return ptr;
}

EXPORT void managed_free(void* ptr, size_t size, int device, void* stream) {
  // Memory allocated with cudaMallocManaged should be released with cudaFree.
  (void)size; (void)device; (void)stream;
  if (ptr) cudaFree(ptr);
}
"""

# Avoid linting errors.
has_unified_memory = False
_alloc = None

# Build the .so upon import; this avoids issues.
if _has_mem_pool:
    _extra_ldflags = ["-lcudart"]
    if CUDA_HOME:
        _cuda_lib = os.path.join(CUDA_HOME, "lib64")
        if os.path.isdir(_cuda_lib):
            _extra_ldflags = [f"-L{_cuda_lib}", "-lcudart"]
    try:
        _mod = load_inline(
            name="managed_alloc_runtime",
            cpp_sources=[_mempool_c_src],
            functions=[],
            with_cuda=True,
            extra_ldflags=_extra_ldflags,
            verbose=False,
        )
        _so_path = Path(_mod.__file__).as_posix()
        _alloc = CUDAPluggableAllocator(_so_path, "managed_malloc", "managed_free").allocator()
        has_unified_memory = True
    except (RuntimeError, ImportError, OSError):
        warnings.warn("Failed to create unified memory mempool.")


def create_unified_mempool():
    """Create a unified memory mempool using CUDA managed memory."""
    assert has_unified_memory
    return MemPool(allocator=_alloc)



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.

""" Common strategies. """

import logging
import os
from pathlib import Path
from typing import Union

import torch

from megatron.core.dist_checkpointing.mapping import ShardedStateDict, StateDict
from megatron.core.dist_checkpointing.strategies.base import (
    SaveCommonStrategy,
    StrategyAction,
    register_default_strategy,
)
from megatron.core.msc_utils import MultiStorageClientFeature

from ..dict_utils import dict_list_map_inplace, nested_values
from ..mapping import CheckpointingException, ShardedObject, is_main_replica
from ..strategies.base import LoadCommonStrategy

COMMON_STATE_FNAME = 'common.pt'

logger = logging.getLogger(__name__)


def register_default_common_strategies():
    """Register default common strategies."""
    register_default_strategy(StrategyAction.LOAD_COMMON, 'torch', 1, TorchCommonLoadStrategy())
    register_default_strategy(
        StrategyAction.SAVE_COMMON, 'torch', 1, TorchCommonSaveStrategy('torch', 1)
    )


class TorchCommonSaveStrategy(SaveCommonStrategy):
    """Common save strategy leveraging native torch save/load."""

    def save_common(self, common_state_dict: StateDict, checkpoint_dir: Union[str, Path]):
        """Save common part of the state dict."""
        if torch.distributed.get_rank() == 0:
            path = os.path.join(checkpoint_dir, COMMON_STATE_FNAME)
            if MultiStorageClientFeature.is_enabled():
                msc = MultiStorageClientFeature.import_package()
                msc.torch.save(common_state_dict, path)
            else:
                torch.save(common_state_dict, path)

    def save_sharded_objects(
        self, sharded_objects_state_dict: ShardedStateDict, checkpoint_dir: Union[str, Path]
    ):
        """Save sharded objects from the state dict."""
        for sh_obj in nested_values(sharded_objects_state_dict):
            if is_main_replica(sh_obj.replica_id):
                save_path = os.path.join(checkpoint_dir, f"{sh_obj.unique_key}.pt")
                parent_dir = os.path.dirname(save_path)
                if MultiStorageClientFeature.is_enabled():
                    msc = MultiStorageClientFeature.import_package()
                    msc.os.makedirs(parent_dir, exist_ok=True)
                    msc.torch.save(sh_obj.data, save_path)
                else:
                    os.makedirs(parent_dir, exist_ok=True)
                    torch.save(sh_obj.data, save_path)

    def can_handle_sharded_objects(self):
        """This strategy can handle ShardedObjects."""
        return True


class TorchCommonLoadStrategy(LoadCommonStrategy):
    """Common load strategy leveraging native torch save/load."""

    def load_common(self, checkpoint_dir: Union[str, Path]):
        """Load common (non-sharded) objects state dict from the checkpoint.

        Args:
            checkpoint_dir (Union[str, Path]): checkpoint directory

        Returns:
            StateDict: state dict with non-sharded objects from the checkpoint
        """
        load_path = os.path.join(checkpoint_dir, COMMON_STATE_FNAME)
        try:
            if MultiStorageClientFeature.is_enabled():
                msc = MultiStorageClientFeature.import_package()
                return msc.torch.load(load_path, map_location='cpu')
            else:
                return torch.load(load_path, map_location='cpu')
        except FileNotFoundError as e:
            err_msg = f'Common file {load_path} does not exist'
            if MultiStorageClientFeature.is_enabled():
                msc = MultiStorageClientFeature.import_package()
                ckpt_files = [f.name for f in msc.Path(checkpoint_dir).iterdir()]
            else:
                ckpt_files = [f.name for f in checkpoint_dir.iterdir()]
            logger.debug(f'{err_msg}. Checkpoint directory content: {ckpt_files}')
            raise CheckpointingException(err_msg) from e

    def load_sharded_objects(
        self, sharded_objects_state_dict: ShardedStateDict, checkpoint_dir: Union[str, Path]
    ):
        """Replaces all ShardedObject from a given state dict with values loaded from the
        checkpoint.

        Args:
            sharded_objects_state_dict (ShardedStateDict):
                sharded state dict defining what objects should be loaded.
            checkpoint_dir (Union[str, Path]): checkpoint directory

        Returns:
            None: sharded state dict is modified in place
        """

        def load_sharded_object(sh_obj: ShardedObject):
            sh_obj.data = None
            load_path = os.path.join(checkpoint_dir, f'{sh_obj.unique_key}.pt')
            try:
                if MultiStorageClientFeature.is_enabled():
                    msc = MultiStorageClientFeature.import_package()
                    loaded_obj = msc.torch.load(load_path)
                else:
                    loaded_obj = torch.load(load_path)
            except FileNotFoundError as e:
                # Backward compatible logic: previously the save format was incorrect
                base, _ = os.path.splitext(sh_obj.unique_key)
                old_load_path = os.path.join(checkpoint_dir, f"{base}.pt")
                try:
                    if MultiStorageClientFeature.is_enabled():
                        msc = MultiStorageClientFeature.import_package()
                        loaded_obj = msc.torch.load(old_load_path)
                    else:
                        loaded_obj = torch.load(old_load_path)
                except FileNotFoundError:
                    err_msg = f'Object shard {load_path} not found'
                    obj_subdir = os.path.join(checkpoint_dir, sh_obj.key)
                    if os.path.exists(obj_subdir):
                        obj_files = os.listdir(obj_subdir)
                        logger.debug(
                            f'{err_msg}. Object {sh_obj.key} directory content: {obj_files}'
                        )
                    else:
                        ckpt_files = os.listdir(checkpoint_dir)
                        logger.debug(
                            f'{err_msg}. Object {sh_obj.key} directory does not exist. Checkpoint'
                            f' directory content: {ckpt_files}'
                        )
                    raise CheckpointingException(err_msg) from e
            return loaded_obj

        return dict_list_map_inplace(load_sharded_object, sharded_objects_state_dict)

    def load_sharded_metadata(self, checkpoint_dir: Union[str, Path]) -> ShardedStateDict:
        if MultiStorageClientFeature.is_enabled():
            msc = MultiStorageClientFeature.import_package()
            checkpoint_dir = msc.Path(checkpoint_dir)
        else:
            checkpoint_dir = Path(checkpoint_dir)

        sharded_metadata = {}
        for subdir in checkpoint_dir.iterdir():
            if not subdir.is_dir():
                continue
            shard_files = list(subdir.glob('shard_*.pt'))
            if not shard_files:
                continue
            sh_objs = []
            for shard_file in shard_files:
                full_key = f'{subdir.name}/{shard_file.stem}'
                sh_objs.append(ShardedObject.empty_from_unique_key(full_key))

            # This is a backward-compatibility fix, where the last global shape is missing in the
            # name
            if sh_objs[0].global_shape[-1] < 0:
                max_last_offset = max(map(lambda sh_obj: sh_obj.global_offset[-1], sh_objs))
                for sh_obj in sh_objs:
                    sh_obj.global_shape = (*sh_obj.global_shape[:-1], max_last_offset + 1)

            # Update the sharded state dict
            for sh_obj in sh_objs:
                sharded_metadata[sh_obj.unique_key] = sh_obj
        return sharded_metadata

    @property
    def can_handle_sharded_objects(self):
        """This strategy can handle ShardedObjects."""
        return True

    def check_backend_compatibility(self, loaded_version):
        pass

    def check_version_compatibility(self, loaded_version):
        pass



# Copyright (c) 2025, NVIDIA CORPORATION. All rights reserved.

from enum import Enum


class Headers(Enum):
    """
    Enum representing headers used for communication with the inference-coordinator.
    """

    CONNECT = 0
    ACK = 1
    SUBMIT_REQUEST = 2
    ENGINE_REPLY = 3
    PAUSE = 4
    UNPAUSE = 5
    STOP = 6



# Copyright (c) 2024, NVIDIA CORPORATION.  All rights reserved.

"""Validate an index's data.

This module contains functionality for checking for bitwise equality across code
changes. The training and adding steps of index construction can be validated
separately. The following high-level checks are supported:

  - Training: Validate that saved training embeddings are bitwise equal with a
      sample set of freshly computed embeddings. (*Note*:
      `--no-retro-index-delete-training-embeddings` must be used.)
  - Adding: Validate that the saved encodings are bitwise equal with a sample of
      sample set of freshly computed encodings. (*Note*:
      `--no-retro-index-delete-added-codes` must be used.)
"""

import numpy as np
import torch
from torch.utils.data import Subset

from megatron.core.datasets.retro.config import RetroPreprocessingConfig
from megatron.core.datasets.retro.utils import get_blocks_by_rank, log_retro_rank_0

from .build import get_text_dataset_for_adding, get_text_dataset_for_training
from .factory import IndexFactory
from .utils import get_added_codes_dir, get_training_data_block_dir

try:
    import h5py

    HAVE_H5PY = True
except ImportError:
    HAVE_H5PY = False


##################################################
# Validate trained index.
##################################################


def validate_training_embeddings(config: RetroPreprocessingConfig) -> None:
    """Validate training embeddings.

    Steps:
    - Randomly sample subset of text dataset blocks.
    - Embed each block.
    - Compare against saved embeddings.

    Args:
        config (RetroPreprocessingConfig): Retro preprocessing config.
    """

    if not HAVE_H5PY:
        raise ImportError(
            "h5py is required to use the validate_training_embeddings function. "
            "Please install h5py."
        )

    # Training text dataset.
    text_dataset = get_text_dataset_for_training(config)

    # Sample existing blocks.
    blocks = get_blocks_by_rank(
        dirname=get_training_data_block_dir(config),
        n_samples=len(text_dataset),
        block_size=config.retro_block_size,
        validate=None,
        sample=config.retro_task_validate,
    )

    assert blocks.n_missing_world == 0

    # Embed & validate blocks.
    embedder = config.retro_bert_embedders.mem
    for block_idx, block in enumerate(blocks.existing):
        # Missing block lists are extended with None to have equal-length
        # lists. Skip the Nones.
        if block is not None:
            # Progress. (*note*: move world progress to here.)
            log_retro_rank_0(
                "embed training block %d / %d ... %s."
                % (block_idx, len(blocks.existing), block["path"])
            )

            # Load existing block embeddings.
            with h5py.File(block["path"]) as f:
                existing_embeddings = np.copy(f["data"])

            # Embed block.
            sub_dataset = Subset(text_dataset, range(*block["range"]))
            embeddings = embedder.embed_text_dataset(sub_dataset, "train")

            # Check equality.
            log_retro_rank_0(" > validate.")
            assert np.array_equal(existing_embeddings, embeddings)

        # Synchronize progress across all ranks. (for easier observation)
        log_retro_rank_0(" > waiting for other ranks to finish block.")
        torch.distributed.barrier()

    log_retro_rank_0(" > finished validating training embeddings.")


##################################################
# Validate filled index.
##################################################


def validate_added_encodings(config: RetroPreprocessingConfig) -> None:
    """Validate added encodings.

    Steps:
    - Randomly sample subset of text dataset blocks.
    - Encode each block.
    - Compare against saved encodings.

    Args:
        config (RetroPreprocessingConfig): Retro preprocessing config.
    """

    # Index.
    index = IndexFactory.get_index(config.retro_index_type)
    inner_index = index.get_empty_index(config)

    # Text dataset.
    text_dataset = get_text_dataset_for_adding(config)

    # Sample existing blocks.
    def validate(f: h5py.File) -> None:
        """Validation method for validating encoding blocks.

        Args:
            f (h5py.File): File with block of encodings.
        """
        assert len(f["data"].shape) == 2

    blocks = get_blocks_by_rank(
        dirname=get_added_codes_dir(config),
        n_samples=len(text_dataset),
        block_size=config.retro_block_size,
        validate=validate,
        sample=config.retro_task_validate,
    )

    assert blocks.n_missing_world == 0

    # Encode and validate blocks.
    embedder = config.retro_bert_embedders.mem
    for block_idx, block in enumerate(blocks.existing):
        if block is not None:
            # Progress.
            log_retro_rank_0(
                "encode block %d / %d ... %s." % (block_idx, len(blocks.existing), block["path"])
            )

            # Load existing codes.
            with h5py.File(block["path"]) as f:
                existing_codes = np.copy(f["data"])

            # Encode block.
            embeddings, codes = index.encode_block(inner_index, embedder, text_dataset, block)

            # Check equality.
            log_retro_rank_0(" > validate.")
            assert np.array_equal(existing_codes, codes)

        # Synchronize progress across all ranks. (for easier observation)
        log_retro_rank_0(" > waiting for other ranks to finish block.")
        torch.distributed.barrier()

    log_retro_rank_0(" > finished validating added encodings.")


##################################################
# Validate index (trained + filled).
##################################################


def validate_index(config: RetroPreprocessingConfig) -> None:
    """Validate index.

    Validating index involves sequentially running stages above:
    - Validate trained index.
    - Validate filled index.

    Args:
        config (RetroPreprocessingConfig): Retro preprocessing config.
    """

    # Validate training embeddings.
    validate_training_embeddings(config)

    # Validate added codes.
    validate_added_encodings(config)



# Copyright (c) 2024, NVIDIA CORPORATION. All rights reserved.
import functools
import time
import typing
import warnings
from collections import OrderedDict
from typing import Dict, Optional, Type, Union

import torch

from megatron.core.inference.async_stream import AsyncStream
from megatron.core.inference.inference_request import InferenceRequest, Status
from megatron.core.inference.sampling_params import SamplingParams
from megatron.core.inference.utils import Counter


class Scheduler:
    """Scheduler for handling requests to inference engine

    This class is responsible for handing of all the incomign requests

    Args:
        max_batch_size (int): The max batch size that we can pass to the
            inference engine at a time.
        request_type (InferenceRequest): The class to use for instantiating new requests.
    """

    def __init__(self, max_batch_size):
        self.max_batch_size = max_batch_size
        self.requests: Dict[int, InferenceRequest] = OrderedDict()
        self.streams: Dict[int, AsyncStream] = OrderedDict()
        self.active_request_pool: Dict[int, InferenceRequest] = OrderedDict()
        self.waiting_request_pool: Dict[int, InferenceRequest] = OrderedDict()
        self.completed_request_pool: Dict[int, InferenceRequest] = OrderedDict()
        self.request_counter = Counter()

    def get_new_request_id(self) -> int:
        """Gets a new request id"""
        request_id = int(next(self.request_counter))
        return request_id

    def add_request(
        self,
        prompt: Optional[str] = None,
        prompt_tokens: Optional[torch.Tensor] = None,
        encoder_prompt: Optional[str] = None,
        sampling_params: Optional[SamplingParams] = None,
        arrival_time: Optional[float] = None,
        streaming: bool = False,
        inference_request: Optional[InferenceRequest] = None,
        *,
        inference_parameters: Optional[SamplingParams] = None,
    ) -> int:
        """Add an incoming request

        This method will add the request to either the active pool or the waiting pool
        depending on the batch size.

        Args:
            prompt (str): Input prompt string
            prompt_tokens (torch.Tensor): A torch tensor having the input prompts tokenized
            encoder_prompt (str): Encoder input string
            sampling_params (SamplingParams): The sampling parameters
            arrival_time (float, optional): The incoming request time. Defaults to None.
            streaming (bool, optional): Whether to asynchronously stream tokens for this request.
            inference_request (InferenceRequest, optional): A fully constructed request.
                Defaults to None.

        Returns:
            The request_id for the new request.
        """
        status = (
            Status.ACTIVE_BUT_NOT_GENERATING_TOKENS
            if len(self.active_request_pool) < self.max_batch_size
            else Status.WAITING_IN_QUEUE
        )

        # Deprecation warning for `inference_parameters`.
        if inference_parameters is not None:
            warnings.warn(
                "`inference_parameters` has been renamed to `sampling_params`, and the "
                "previous name will be removed in `megatron-core` 0.13."
            )
            if sampling_params is None:
                sampling_params = inference_parameters

        if inference_request is None:
            assert prompt is not None
            assert prompt_tokens is not None

            request_id = self.get_new_request_id()

            if arrival_time is None:
                arrival_time = time.time()

            inference_request = InferenceRequest(
                request_id=request_id,
                prompt=prompt,
                sampling_params=sampling_params,
                arrival_time=arrival_time,
                prompt_tokens=prompt_tokens,
                status=status,
                encoder_prompt=encoder_prompt,
            )
        else:
            request_id = inference_request.request_id
            inference_request.status = status
            if inference_request.arrival_time is None:
                inference_request.arrival_time = time.time()

        self.requests[request_id] = inference_request

        if streaming:
            abort_request = functools.partial(self.abort_request, request_id=request_id)
            self.streams[request_id] = AsyncStream(request_id, abort_request)

        if status == status.ACTIVE_BUT_NOT_GENERATING_TOKENS:
            self.active_request_pool[request_id] = inference_request
        else:
            self.waiting_request_pool[request_id] = inference_request

        return request_id

    def num_requests_pending(self) -> int:
        """Get the number of requests pending.

        This method returns the number of active + waiting requests.
        """
        return len(self.active_request_pool) + len(self.waiting_request_pool)

    def have_requests_pending(self) -> bool:
        """Method to check if there are requests pending.

        This method returns False only when there are no active requests or waiting requests.
        """
        return self.num_requests_pending() > 0

    def add_earliest_waiting_request_to_active_pool(self):
        """Utility to add the waiting request to active pool

        This method will add the earliest request (FIFO) that is in the waiting request
        pool to the active request pool.
        """
        assert (
            len(self.active_request_pool) < self.max_batch_size
        ), "Active request pool is already full. Cant add any more requests"
        if len(self.waiting_request_pool) > 0:
            (earliest_waiting_request_request_id, earliest_waiting_request) = (
                self.waiting_request_pool.popitem(last=False)
            )
            earliest_waiting_request.status = Status.ACTIVE_BUT_NOT_GENERATING_TOKENS
            self.active_request_pool[earliest_waiting_request_request_id] = earliest_waiting_request

    def update_requests_pools(
        self, result_dict: Optional[typing.OrderedDict[int, InferenceRequest]] = None
    ):
        """Update request pool status

        This method will full up the active request pool, if it has less than max batch size
        elements from the waiting request pool.
        If provided with a request dict, it will put the completed requests into the completed
        request pool and add waiting request into active pool.

        Args:
            result (typing.OrderedDict[int, InferenceRequest], optional): The result returned
                by the engine. A dictionary with keys as the request ids, and values as the
                requests. Defaults to None.
        """
        for result_request_id in list(result_dict.keys()):
            active_request = self.active_request_pool[result_request_id]

            # If a request has completed put it into the completed request pool.
            if active_request.status == Status.COMPLETED:
                completed_request = self.active_request_pool.pop(result_request_id)
                self.completed_request_pool[result_request_id] = completed_request

        # If the active request pool is not full, add waiting requests in FIFO order
        while (
            len(self.active_request_pool) < self.max_batch_size
            and len(self.waiting_request_pool) > 0
        ):
            self.add_earliest_waiting_request_to_active_pool()

    def abort_request(
        self,
        request_id: int,
        *,
        exception: Optional[Union[BaseException, Type[BaseException]]] = None,
    ):
        """Cancels the given request"""
        stream = self.streams.get(request_id, None)
        if stream is not None:
            stream.finish(exception=exception)



# Copyright (c) 2024, NVIDIA CORPORATION. All rights reserved.
# Copyright 2018-2020 Philippe Tillet
# Copyright 2020-2022 OpenAI

# Some of this code was adopted from https://github.com/triton-lang/triton
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import uuid
from pathlib import Path

try:
    from triton import __version__ as triton_version
    from triton.runtime.cache import FileCacheManager
except ImportError:
    raise ImportError("triton is required by the Mamba model but cannot be imported")


def _version_no_greater_than(version, version_limit):
    major, minor, _ = map(int, version.split('.'))
    limit_major, limit_minor = map(int, version_limit.split('.'))
    return major < limit_major or (major == limit_major and minor <= limit_minor)


def default_cache_dir():
    """Provides a default path for the Triton cache directory."""
    return os.path.join(Path.home(), ".triton", "cache")


class ParallelFileCacheManager(FileCacheManager):
    """
    This patched version of ParallelFileCacheManager prevents errors related
    to the builing of the Triton compiler cache when the number of model
    parallel ranks is greater than one, including when certain types of file
    system are used (such as Lustre).

    Usage:
    export TRITON_CACHE_DIR=<chosen-cache-location>
    export TRITON_CACHE_MANAGER=megatron.core.ssm.triton_cache_manager:ParallelFileCacheManager

    This patch implements the changes in the following two Triton project pull
    requests:
    1. https://github.com/triton-lang/triton/pull/3544
    2. https://github.com/triton-lang/triton/pull/4295

    The above changes will probably be included in Triton release version 3.2,
    making this patch no longer necessary.
    """

    def put(self, data, filename, binary=True) -> str:
        """A patched version of put, implementing PR 3544 and PR 4295."""
        patch_limit = '3.1'
        assert _version_no_greater_than(triton_version, patch_limit), (
            "Assertion failed: ParallelFileCacheManager patch should not be "
            f"used beyond Triton version {patch_limit}."
        )
        if not self.cache_dir:
            raise RuntimeError("Could not create or locate cache dir")
        binary = isinstance(data, bytes)
        if not binary:
            data = str(data)
        assert self.lock_path is not None
        filepath = self._make_path(filename)
        # Random ID to avoid any collisions
        rnd_id = str(uuid.uuid4())
        # we use the PID in case a bunch of these around so we can see what PID made it
        pid = os.getpid()
        # use temp dir to be robust against program interruptions
        temp_dir = os.path.join(self.cache_dir, f"tmp.pid_{pid}_{rnd_id}")
        os.makedirs(temp_dir, exist_ok=True)
        temp_path = os.path.join(temp_dir, filename)

        mode = "wb" if binary else "w"
        with open(temp_path, mode) as f:
            f.write(data)
        # Replace is guaranteed to be atomic on POSIX systems if it succeeds
        # so filepath cannot see a partial write
        os.replace(temp_path, filepath)
        os.removedirs(temp_dir)
        return filepath
